from __future__ import annotations
from typing import TYPE_CHECKING, Any, List, Union, Sequence, Optional, Iterable, Tuple, Callable, Type, cast
from types import MethodType
from argparse import Action, SUPPRESS, ArgumentParser, Namespace, HelpFormatter
from textwrap import fill as textwrap_fill

if TYPE_CHECKING:  # pragma: no cover
    from typing_extensions import Self


__all__ = [
    'MarkdownHelpAction',
    'MarkdownFormatter',
]


class MarkdownHelpAction(Action):
    def __init__(
        self,
        option_strings: List[str],
        dest: str = SUPPRESS,
        default: str = SUPPRESS,
        help: str = SUPPRESS,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            option_strings=option_strings,
            dest=dest,
            default=default,
            help=help,
            nargs=0,
            **kwargs,
        )

    def __call__(
        self,
        parser: ArgumentParser,
        namespace: Namespace,
        values: Union[str, Sequence[Any], None],
        option_string: Optional[str] = None,
    ) -> None:
        self.print_help(parser)

        parser.exit()

    def print_help(self, parser: ArgumentParser, level: int = 0) -> None:
        def format_help_markdown(self: ArgumentParser) -> str:
            formatter = self._get_formatter()

            # description -- in markdown, should come before usage
            formatter.add_text('\n')
            formatter.add_text(self.description)

            # usage
            formatter.add_text('\n')
            formatter.add_usage(self.usage, self._actions,
                                self._mutually_exclusive_groups)

            # XXX: formatter.add_text(self.description) -- used to be here

            # positionals, optionals and user-defined groups
            for action_group in self._action_groups:
                formatter.start_section(action_group.title)
                formatter.add_text(action_group.description)
                formatter.add_arguments(action_group._group_actions)
                formatter.end_section()

            # epilog
            formatter.add_text(self.epilog)

            # determine help from format above
            return formatter.format_help()

        # <!-- monkey patch our parser
        # switch format_help, so that stuff comes in an order that makes more sense in markdown
        setattr(parser, 'format_help', MethodType(format_help_markdown, parser))
        # switch formatter class so we'll get markdown
        setattr(parser, 'formatter_class', MarkdownFormatter.factory(level))
        # -->

        parser.print_help()

        # check if the parser has a subparser, so we can generate its
        # help in markdown as well
        _subparsers = getattr(parser, '_subparsers', None)
        if _subparsers is not None:
            for subparsers in _subparsers._group_actions:
                for subparser in subparsers.choices.values():
                    self.print_help(subparser, level=level + 1)


class MarkdownFormatter(HelpFormatter):
    level: int

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._root_section = self._MarkdownSection(self, None)
        self._current_section = self._root_section

    @staticmethod
    def factory(level: int) -> Type['MarkdownFormatter']:
        return type('MarkdownFormatterInstance', (MarkdownFormatter,), {'level': level})

    class _MarkdownSection(HelpFormatter._Section):
        def __init__(self, formatter: 'MarkdownFormatter', parent: Optional[Self], heading: Optional[str] = None) -> None:
            self.formatter = formatter
            self.parent = parent
            self.heading = heading
            self.items: List[Tuple[Callable[..., str], Iterable[Any]]] = []

        def format_help(self) -> str:
            # format the indented section
            if self.parent is not None:
                self.formatter._indent()
            join = self.formatter._join_parts
            helps: List[str] = []

            # only one table header per section
            print_table_headers = True
            for func, args in self.items:
                item_help_text = func(*args)
                name = getattr(func, '__name__', repr(func))

                # we need to fix headers for argument tables
                if name == '_format_action':
                    if print_table_headers and len(item_help_text) > 0:
                        helps.extend([
                            '\n',
                            '| argument | default | help |',
                            '\n',
                            '| -------- | ------- | ---- |',
                            '\n',
                        ])
                        print_table_headers = False

                helps.append(item_help_text)

            item_help = join(helps)

            if self.parent is not None:
                self.formatter._dedent()

            # return nothing if the section was empty
            if not item_help:
                return ''

            # add the heading if the section was non-empty
            if self.heading is not SUPPRESS and self.heading is not None:
                current_indent = self.formatter._current_indent
                heading = '%*s%s\n' % (current_indent, '', self.heading)

                # increase header if we're in a subparser
                assert isinstance(self.formatter, MarkdownFormatter)
                if self.formatter.level > 0:
                    # a bit hackish, to get a line break when adding a subparsers help
                    if self.parent is None:
                        print('')

                    heading = f'#{heading}'
            else:
                heading = ''

            # join the section-initial newline, the heading and the help
            return join(['\n', heading, item_help, '\n'])

    @property
    def current_level(self) -> int:
        return self.level + 1

    def _format_usage(self, *args: Any, **kwargs: Any) -> str:
        # remove last argument, which is prefix, we are going to set it
        args = args[:-1]
        args += ('', )
        usage_text = super()._format_usage(*args, **kwargs)

        # wrap usage text in a markdown code block, with bash syntax
        return '\n'.join([
            '',
            f'{"#" * self.current_level}## Usage',
            '',
            '```bash',
            usage_text.strip(),
            '```',
            '',
        ])

    def format_help(self) -> str:
        heading = f'{"#" * self.current_level} `{self._prog}`'
        self._root_section.heading = heading
        return super().format_help()

    def _format_text(self, text: str) -> str:
        if '%(prog)' in text:
            text = text % dict(prog=self._prog)

        if len(text.strip()) > 0:
            lines: List[str] = []
            for line in text.split('\n'):
                line = textwrap_fill(line, 120)
                lines.append(line)
            text = '\n'.join(lines)

        return text

    def start_section(self, heading: Optional[str]) -> None:
        if heading is not None:
            heading = f'{heading[0].upper()}{heading[1:]}'  # first letter in first words to upper case
            heading = f'{"#" * self.current_level}# {heading}'

        self._indent()
        section = self._MarkdownSection(self, cast(MarkdownFormatter._MarkdownSection, self._current_section), heading)
        self._add_item(section.format_help, [])
        self._current_section = section

    def _format_action(self, action: Action) -> str:
        # do not include -h/--help or --md-help in the markdown
        # help
        if 'help' in action.dest or action.dest == SUPPRESS:
            return ''

        lines: List[str] = []

        if action.help is not None:
            expanded_help = self._expand_help(action)
            help_text = self._split_lines(expanded_help, 80)
        else:
            help_text = ['']

        argument = ', '.join(action.option_strings) if getattr(action, 'option_strings', None) is not None and len(action.option_strings) > 0 else action.dest
        default = f'`{action.default}`' if action.default is not None else ''
        help = '<br/>'.join(help_text)

        # format arguments as a markdown table row
        lines.extend([f'| `{argument}` | {default} | {help} |', ''])

        return '\n'.join(lines)



import sys

from typing import Any, Dict, List, Union, Sequence, Optional, cast
from argparse import (
    ArgumentParser,
    ArgumentError,
    Action,
    Namespace,
    _SubParsersAction,
    _StoreConstAction,
    _AppendAction,
    _StoreAction,
    SUPPRESS,
)
from os import path

from .types import BashCompletionTypes

__all__ = [
    'BashCompletionTypes',
    'BashCompletionAction',
    'BashCompleteAction',
    'hook',
]


class BashCompletionAction(Action):
    def __init__(
        self,
        option_strings: List[str],
        dest: str = SUPPRESS,
        default: str = SUPPRESS,
        help: str = SUPPRESS,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            option_strings=option_strings,
            dest=dest,
            default=default,
            help=help,
            nargs=0,
            **kwargs,
        )

    def __call__(
        self,
        parser: ArgumentParser,
        namespace: Namespace,
        values: Union[str, Sequence[Any], None],
        option_string: Optional[str] = None,
    ) -> None:
        file_directory = path.dirname(__file__)
        with open(path.join(file_directory, 'bashcompletion.bash'), encoding='utf-8') as fd:
            print(fd.read().replace('bashcompletion_template', parser.prog))

        parser.exit()


class BashCompleteAction(Action):
    def __init__(
        self,
        option_strings: List[str],
        dest: str = SUPPRESS,
        default: str = SUPPRESS,
        help: str = SUPPRESS,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            option_strings=option_strings,
            dest=dest,
            default=default,
            help=help,
            nargs=None,
            **kwargs,
        )

    def get_suggestions(self, parser: ArgumentParser) -> Dict[str, Union[str, Action]]:
        suggestions: Dict[str, Union[str, Action]] = {}

        for action in parser._actions:
            if isinstance(action, (BashCompleteAction, BashCompletionAction, )) or (SUPPRESS in [action.help, action.default] and action.dest != 'help'):
                continue
            elif isinstance(action, _SubParsersAction):
                suggestions.update({key: action for key in action.choices.keys()})
            else:
                if len(action.option_strings) > 0:
                    suggestions.update({key: action for key in action.option_strings})
                else:
                    suggestions.update({action.dest: action})

        return suggestions

    def get_exclusive_suggestions(self, parser: ArgumentParser) -> Dict[str, List[str]]:
        exclusive_suggestions: Dict[str, List[str]] = {}
        for group in parser._mutually_exclusive_groups:
            exclusives: List[str] = []
            for action in group._group_actions:
                for option in action.option_strings:
                    exclusives.append(option)

            for exclusive in exclusives:
                exclusives_to: List[str] = []
                for exclusive_to in filter(lambda x: x != exclusive, exclusives):
                    exclusives_to.append(exclusive_to)

                exclusive_suggestions.update({exclusive: exclusives_to})

        return exclusive_suggestions

    def get_provided_options(self, prog: str, values: Union[str, Sequence[Any], None]) -> List[str]:
        options: List[str] = []

        if isinstance(values, str):
            options = [value for value in values.replace(f'{prog}', '').split(' ') if len(value.strip()) > 0]
        elif isinstance(values, Sequence):
            options = [str(value) for value in values if len(str(value).strip()) > 0 and prog not in value]

        return options

    def remove_completed(self, provided_options: List[str], suggestions: Dict[str, Union[str, Action]], exclusive_suggestions: Dict[str, List[str]]) -> List[str]:
        if len(provided_options) <= 1:
            return provided_options

        filtered_options: List[str] = []
        skip: bool = False
        concat: bool = False

        for index, option in enumerate(provided_options):
            add_next = False
            remove_suggestion = False

            if concat:
                concat = False
                option = '{} {}'.format(provided_options[index - 1], option)

            if len(option) < 1 or skip:
                skip = False
                continue

            suggestion = suggestions.get(option, None)

            if suggestion is not None:
                if isinstance(suggestion, _AppendAction):
                    # skip next supplied value, since it is part of the option
                    # also, _AppendAction can be specified more than once
                    if len(provided_options) > index + 1 and not provided_options[index + 1].strip().startswith('-'):
                        skip = True
                        continue
                elif isinstance(suggestion, _StoreAction):
                    if index == len(provided_options) - 1:
                        pass
                    elif index == len(provided_options) - 2:  # beginning of value should also be added
                        add_next = True
                    else:  # completed, do not add
                        remove_suggestion = True
                else:
                    remove_suggestion = True

                if remove_suggestion and isinstance(suggestion, Action):
                    # remove all other, completed, options from suggestion
                    for suggestion_option in suggestion.option_strings:
                        if suggestion_option in suggestions:
                            del suggestions[suggestion_option]

                        # remove options that are mutually exclusive to completed option
                        exclusive_removes = exclusive_suggestions.get(suggestion_option, [])
                        for exclusive_option in exclusive_removes:
                            if exclusive_option in suggestions:
                                del suggestions[exclusive_option]
                    continue
            elif not any([suggested_option.startswith(option) for suggested_option in suggestions.keys()]):  # could be values for an option
                remove = True
                if option.endswith('\\') and sys.platform != 'win32':
                    concat = True
                    continue

                for suggestion in suggestions.values():
                    if isinstance(suggestion, Action) and len(suggestion.option_strings) == 0:
                        if isinstance(suggestion.type, BashCompletionTypes.File):
                            file_suggestions = cast(BashCompletionTypes.File, suggestion.type).list_files(option)  # type: ignore
                            for file in file_suggestions.keys():
                                if file.startswith(option):
                                    remove = False
                                    break

                if remove:
                    continue

            filtered_options.append(option)
            if add_next:
                filtered_options.append(provided_options[index + 1])
                skip = True

        return filtered_options

    def filter_suggestions(self, provided_options: List[str], suggestions: Dict[str, Union[str, Action]]) -> Dict[str, Union[str, Action]]:
        if len(provided_options) < 1:
            return suggestions

        filtered_suggestions: Dict[str, Union[str, Action]] = {}
        for option in provided_options:
            for option_suggestion, suggestion in suggestions.items():
                if option_suggestion.startswith(option) or (
                    not option.startswith('-')
                    and isinstance(suggestion, Action)
                    and len(suggestion.option_strings) == 0
                    and not isinstance(suggestion, _SubParsersAction)
                ):
                    filtered_suggestions.update({option_suggestion: suggestion})

        return filtered_suggestions

    def __call__(
        self,
        parser: ArgumentParser,
        namespace: Namespace,
        values: Union[str, Sequence[Any], None],
        option_string: Optional[str] = None,
    ) -> None:
        all_suggestions: Dict[str, Union[str, Action]] = {}
        provided_options = self.get_provided_options(parser.prog, values)
        suggestions = self.get_suggestions(parser)

        if '-h' in provided_options or '--help' in provided_options:
            print('')
            parser.exit()

        # map options that are only allowed by it own
        exclusive_suggestions = self.get_exclusive_suggestions(parser)

        # check for positional arguments (e.g. no option string)

        # remove any completed values, or exclusive suggestions
        provided_options = self.remove_completed(provided_options, suggestions, exclusive_suggestions)

        # all, remaining, suggestions -- completed has been removed
        all_suggestions = suggestions.copy()

        # only return suggestions that matches supplied value
        suggestions = self.filter_suggestions(provided_options, suggestions)

        if len(provided_options) > 0:
            suggestion = suggestions.get(provided_options[0], None)

            if suggestion is not None:
                # based option type
                if isinstance(suggestion, _StoreConstAction):
                    suggestions = all_suggestions
                    for option in suggestion.option_strings:
                        del suggestions[option]
                elif isinstance(suggestion, (_AppendAction, _StoreAction,)):
                    # value for append action has been provided
                    if len(provided_options) == 2:
                        suggestions = all_suggestions
                    else:
                        # no value, supplied, do not suggest anything
                        suggestions = {}

            # based on option value type
            suggestion = all_suggestions.get(provided_options[0], None)
            if isinstance(suggestion, Action) and suggestion.type is not None:
                value = provided_options[-1] if len(provided_options) > 1 else None
                if isinstance(suggestion.type, BashCompletionTypes.File):
                    file_suggestions = cast(BashCompletionTypes.File, suggestion.type).list_files(value)  # type: ignore

                    if not (len(file_suggestions) == 1 and provided_options[-1] in file_suggestions):
                        suggestions = cast(Dict[str, Union[str, Action]], file_suggestions)
                    else:
                        suggestions = all_suggestions
                        for option in suggestion.option_strings:
                            del suggestions[option]
                else:
                    if suggestion.type == str and not isinstance(value, str):
                        suggestions = {}
                    elif suggestion.type == int and (value is None or not value.isnumeric()):
                        suggestions = {}

                if value is not None and isinstance(suggestion, _StoreAction):
                    for option in suggestion.option_strings:
                        if option in suggestions:
                            del suggestions[option]

        # check for positionals
        original_suggestions = suggestions.copy()
        for option, suggestion in original_suggestions.items():
            if (option.startswith('-') and isinstance(suggestion, Action)) or (not option.startswith('-') and isinstance(suggestion, (_SubParsersAction, str,))):
                continue

            del suggestions[option]

            if isinstance(suggestion, Action) and suggestion.type is not None:
                if isinstance(cast(Any, suggestion.type), BashCompletionTypes.File):
                    value = provided_options[-1] if len(provided_options) == 1 and not provided_options[-1].startswith('-') else None

                    if (value is None and len(provided_options) == 0) or (value is not None and len(provided_options) == 1):
                        file_suggestions = cast(BashCompletionTypes.File, suggestion.type).list_files(value)
                        value_type = file_suggestions.get(value, None) if value is not None else None

                        # check if suggestion matching provided option (value) is a directory, and if
                        # provded option (value) does not end with a path separator, it should be added
                        # otherwise it will not be completed correctly
                        if value_type == 'dir' and (value is not None and not value.endswith(path.sep)):
                            value = '{value}{sep}'.format(value=value, sep=path.sep)

                        # only provide further suggestions if matches isn't a completed file path
                        if not (len(file_suggestions) == 1 and value in file_suggestions) and (value_type is None or value_type != 'file'):
                            suggestions.update(file_suggestions)
                        else:
                            suggestions = all_suggestions
                            del suggestions[suggestion.dest]

        print('\n'.join(suggestions.keys()))
        parser.exit()


def hook(parser: ArgumentParser) -> None:
    try:
        parser.add_argument('--bash-complete', action=BashCompleteAction)
    except ArgumentError as e:
        # we've already "hooked" the parser
        if 'conflicting option string: --bash-complete' not in e.message:
            raise e
    except Exception:
        raise
    finally:
        _subparsers = getattr(parser, '_subparsers', None)
        if _subparsers is not None:
            for subparsers in _subparsers._group_actions:
                for subparser in subparsers.choices.values():
                    hook(subparser)



from __future__ import annotations

import re
import sys
import subprocess
import signal as psignal
import logging
import logging.config
import stat
import os

from typing import Optional, List, Set, Union, Dict, Any, Tuple, Callable, Type, ClassVar, cast
from types import TracebackType, FrameType
from os import path, environ
from shutil import which, rmtree
from behave.parser import parse_file as feature_file_parser
from argparse import Namespace as Arguments
from json import loads as jsonloads
from functools import wraps
from packaging import version as versioning
from tempfile import mkdtemp
from hashlib import sha1
from math import ceil
from datetime import datetime, timezone
from dataclasses import dataclass, field
from pathlib import Path
from copy import deepcopy
from collections.abc import Mapping

import requests
import tomli

from behave.model import Scenario
from jinja2 import Template
from progress.spinner import Spinner
from yaml import Dumper

import grizzly_cli


logger = logging.getLogger('grizzly-cli')


class SignalHandler:
    handler: Callable[[int, Optional[FrameType]], None]
    signals: Dict[int, Union[Callable[[int, Optional[FrameType]], Any], int, None]]

    def __init__(self, handler: Callable[[int, Optional[FrameType]], None], signal: int, *signals: int) -> None:
        self.handler = handler
        self.signals = {signal: None}

        if signals is not None and len(signals) > 0:
            for sig in signals:
                self.signals.update({sig: None})

    def __enter__(self) -> None:
        for signal in self.signals.keys():
            self.signals.update({signal: psignal.getsignal(signal)})
            psignal.signal(signal, self.handler)

    def __exit__(self, exc_type: Optional[Type[BaseException]], exc: Optional[BaseException], tb: Optional[TracebackType]) -> bool:
        for signal, handler in self.signals.items():
            psignal.signal(signal, handler)

        return exc is None


@dataclass
class RunCommandResult:
    return_code: int
    abort_timestamp: Optional[datetime] = field(init=False, default=None)
    output: Optional[List[bytes]] = field(init=False, default=None)


def run_command(command: List[str], env: Optional[Dict[str, str]] = None, *, silent: bool = False, verbose: bool = False, spinner: Optional[str] = None) -> RunCommandResult:
    if env is None:
        env = environ.copy()

    if verbose:
        logger.info(f'run_command: {" ".join(command)}')

    process = subprocess.Popen(
        command,
        env=env,
        stderr=subprocess.STDOUT,
        stdout=subprocess.PIPE,
    )

    result = RunCommandResult(return_code=-1)

    if silent:
        result.output = []

    _spinner: Optional[Spinner] = None

    if spinner is not None:  # pragma: no cover
        _spinner = Spinner(f'{spinner} ')

    def sig_handler(signum: int, frame: Optional[FrameType] = None) -> None:  # pragma: no cover
        if result.abort_timestamp is None:
            result.abort_timestamp = datetime.now(timezone.utc)
            process.terminate()

    with SignalHandler(sig_handler, psignal.SIGINT, psignal.SIGTERM):
        try:
            while process.poll() is None:
                stdout = process.stdout
                if stdout is None:
                    break

                if _spinner is not None:  # pragma: no cover
                    _spinner.next()

                output = stdout.readline()
                if not output:
                    break

                if result.output is None:
                    if spinner is None:
                        logger.info(output.decode().rstrip())
                else:
                    result.output.append(output)

            process.terminate()
        except KeyboardInterrupt:
            pass
        finally:
            try:
                process.kill()
            except Exception:
                pass

    process.wait()

    if spinner is not None:  # pragma: no cover
        logger.info('')

    result.return_code = process.returncode

    return result


def get_docker_compose_version() -> Tuple[int, int, int]:  # pragma: no cover
    output = subprocess.getoutput('docker compose version')

    version_line = output.splitlines()[0]

    match = re.match(r'.*version [v]?([1-2]\.[0-9]+\.[0-9]+).*$', version_line)

    if match:
        version = cast(Tuple[int, int, int], tuple([int(part) for part in match.group(1).split('.')]))
    else:
        version = (0, 0, 0,)

    return version


def onerror(func: Callable, path: str, exc_info: Union[
        BaseException,
        Tuple[Type[BaseException], BaseException, Optional[TracebackType]],
]) -> None:  # noqa: ARG001  # pragma: no cover
    """Error handler for shutil.rmtree.

    If the error is due to an access error (read only file)
    it attempts to add write permission and then retries.
    If the error is for another reason it re-raises the error.
    Usage : ``shutil.rmtree(path, onerror=onerror)``
    """
    _path = Path(path)
    # Is the error an access error?
    if not os.access(_path, os.W_OK):
        _path.chmod(stat.S_IWUSR)
        func(path)
    else:
        raise  # pylint: disable=E0704


def rm_rf(path: Union[str, Path], *, missing_ok: bool = False) -> None:
    """Remove the path contents recursively, even if some elements
    are read-only."""
    p = path.as_posix() if isinstance(path, Path) else path

    try:
        if sys.version_info >= (3, 12):
            rmtree(p, onexc=onerror)
        else:  # pragma: no cover
            rmtree(p, onerror=onerror)
    except FileNotFoundError:  # pragma: no cover
        if not missing_ok:
            raise


def get_dependency_versions(local_install: Union[bool, str]) -> Tuple[Tuple[Optional[str], Optional[List[str]]], Optional[str]]:
    grizzly_requirement: Optional[str] = None
    grizzly_requirement_egg: str
    locust_version: Optional[str] = None
    grizzly_version: Optional[str] = None
    grizzly_extras: Optional[List[str]] = None

    args: tuple[str, ...] = (grizzly_cli.EXECUTION_CONTEXT,)
    if isinstance(local_install, str):
        args += (local_install,)
        if not local_install.endswith('requirements.txt'):
            args += ('requirements.txt',)
    else:
        args += ('requirements.txt',)

    project_requirements = path.join(*args)

    try:
        with open(project_requirements, encoding='utf-8') as fd:
            for line in fd.readlines():
                if any([pkg in line for pkg in ['grizzly-loadtester', 'grizzly.git'] if not re.match(r'^([\s]+)?#', line)]):
                    grizzly_requirement = line.strip()
                    break
    except:
        return (None, None,), None

    if grizzly_requirement is None:
        print(f'!! unable to find grizzly dependency in {project_requirements}', file=sys.stderr)
        return ('(unknown)', None, ), '(unknown)'

    # check if it's a repo or not
    if 'git+' in grizzly_requirement:
        if grizzly_requirement.startswith('git+'):
            url, egg_part = grizzly_requirement.rsplit('#', 1)
            _, grizzly_requirement_egg = egg_part.split('=', 1)
        elif grizzly_requirement.index('@') < grizzly_requirement.index('git+'):
            grizzly_requirement_egg, url = grizzly_requirement.split('@', 1)
            grizzly_requirement_egg = grizzly_requirement_egg.strip()
            url = url.strip()
        else:
            print(f'!! unable to find properly formatted grizzly dependency in {project_requirements}', file=sys.stderr)
            return ('(unknown)', None, ), '(unknown)'

        url, branch = url.rsplit('@', 1)
        url = url[4:]  # remove git+
        suffix = sha1(grizzly_requirement.encode('utf-8')).hexdigest()

        # extras_requirement normalization
        egg = grizzly_requirement_egg.replace('[', '__').replace(']', '__').replace(',', '_')

        tmp_workspace = mkdtemp(prefix='grizzly-cli-')
        repo_destination = path.join(tmp_workspace, f'{egg}_{suffix}')

        try:
            rc = subprocess.check_call(
                [
                    'git', 'clone', '--filter=blob:none', '-q',
                    url,
                    repo_destination
                ],
                shell=False,
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
            )

            if rc != 0:
                print(f'!! unable to clone git repo {url}', file=sys.stderr)
                raise RuntimeError()  # abort

            active_branch = branch

            try:
                active_branch = subprocess.check_output(
                    ['git', 'rev-parse', '--abbrev-ref', 'HEAD'],
                    cwd=repo_destination,
                    shell=False,
                    universal_newlines=True,
                ).strip()
                rc = 0
            except subprocess.CalledProcessError as cpe:
                rc = cpe.returncode

            if rc != 0:
                print(f'!! unable to check branch name of HEAD in git repo {url}', file=sys.stderr)
                raise RuntimeError()  # abort

            if active_branch != branch:
                try:
                    git_object_type = subprocess.check_output(
                        ['git', 'cat-file', '-t', branch],
                        cwd=repo_destination,
                        shell=False,
                        universal_newlines=True,
                        stderr=subprocess.STDOUT,
                    ).strip()
                except subprocess.CalledProcessError as cpe:  # pragma: no cover
                    if 'Not a valid object name' in cpe.output:
                        git_object_type = 'branch'  # assume remote branch
                    else:
                        print(f'!! unable to determine git object type for {branch}')
                        raise RuntimeError()

                if git_object_type == 'tag':  # pragma: no cover
                    rc += subprocess.check_call(
                        [
                            'git', 'checkout',
                            f'tags/{branch}',
                            '-b', branch,
                        ],
                        cwd=repo_destination,
                        shell=False,
                        stdout=subprocess.DEVNULL,
                        stderr=subprocess.DEVNULL,
                    )

                    if rc != 0:  # pragma: no cover
                        print(f'!! unable to checkout tag {branch} from git repo {url}', file=sys.stderr)
                        raise RuntimeError()  # abort
                elif git_object_type == 'commit':
                    rc += subprocess.check_call(
                        [
                            'git', 'checkout',
                            branch,
                        ],
                        cwd=repo_destination,
                        shell=False,
                        stdout=subprocess.DEVNULL,
                        stderr=subprocess.DEVNULL,
                    )

                    if rc != 0:  # pragma: no cover
                        print(f'!! unable to checkout commit {branch} from git repo {url}', file=sys.stderr)
                        raise RuntimeError()  # abort
                else:
                    rc += subprocess.check_call(
                        [
                            'git', 'checkout',
                            '-b', branch,
                            '--track', f'origin/{branch}',
                        ],
                        cwd=repo_destination,
                        shell=False,
                        stdout=subprocess.DEVNULL,
                        stderr=subprocess.DEVNULL,
                    )

                    if rc != 0:
                        print(f'!! unable to checkout branch {branch} from git repo {url}', file=sys.stderr)
                        raise RuntimeError()  # abort

            if not path.exists(path.join(repo_destination, 'pyproject.toml')):
                with open(path.join(repo_destination, 'grizzly', '__init__.py'), encoding='utf-8') as fd:
                    version_raw = [line.strip() for line in fd.readlines() if line.strip().startswith('__version__ =')]

                if len(version_raw) != 1:
                    print(f'!! unable to find "__version__" declaration in grizzly/__init__.py from {url}', file=sys.stderr)
                    raise RuntimeError()  # abort

                _, grizzly_version, _ = version_raw[-1].split("'")
            else:
                try:
                    with open(path.join(repo_destination, 'setup.cfg'), encoding='utf-8') as fd:
                        version_raw = [line.strip() for line in fd.readlines() if line.strip().startswith('version = ')]

                    if len(version_raw) != 1:
                        print(f'!! unable to find "version" declaration in setup.cfg from {url}', file=sys.stderr)
                        raise RuntimeError()  # abort

                    _, grizzly_version = version_raw[-1].split(' = ')
                except FileNotFoundError:
                    try:
                        import setuptools_scm  # pylint: disable=unused-import  # noqa: F401  # type: ignore
                    except ModuleNotFoundError:  # pragma: no cover
                        rc = subprocess.check_call([
                            sys.executable,
                            '-m',
                            'pip',
                            'install',
                            'setuptools_scm',
                        ])

                    try:
                        grizzly_version = subprocess.check_output(
                            [
                                sys.executable,
                                '-m',
                                'setuptools_scm',
                            ],
                            shell=False,
                            universal_newlines=True,
                            cwd=repo_destination,
                        ).strip()
                    except subprocess.CalledProcessError:  # pragma: no cover
                        print(f'!! unable to get setuptools_scm version from {url}', file=sys.stderr)
                        raise RuntimeError()  # abort

            if grizzly_version == '0.0.0':
                grizzly_version = '(development)'

            try:
                with open(path.join(repo_destination, 'requirements.txt'), encoding='utf-8') as fd:
                    version_raw = [line.strip() for line in fd.readlines() if line.strip().startswith('locust')]

                if len(version_raw) != 1:
                    print(f'!! unable to find "locust" dependency in requirements.txt from {url}', file=sys.stderr)
                    raise RuntimeError()  # abort

                match = re.match(r'^locust.{2}(.*?)$', version_raw[-1].strip().split(' ')[0])

                if not match:
                    print(f'!! unable to find locust version in "{version_raw[-1].strip()}" specified in requirements.txt from {url}', file=sys.stderr)
                else:
                    locust_version = match.group(1).strip()
            except FileNotFoundError:  # pragma: no cover
                with open(path.join(repo_destination, 'pyproject.toml'), 'rb') as fdt:
                    toml_dict = tomli.load(fdt)
                    dependencies = toml_dict.get('project', {}).get('dependencies', [])
                    for dependency in dependencies:
                        if not dependency.startswith('locust'):
                            continue

                        _, locust_version = dependency.strip().split(' ', 1)

                        break
        except RuntimeError:
            pass
        finally:
            rm_rf(tmp_workspace)
    else:
        response = requests.get(
            'https://pypi.org/pypi/grizzly-loadtester/json'
        )

        if response.status_code != 200:
            print(f'!! unable to get grizzly package information from {response.url} ({response.status_code})', file=sys.stderr)
        else:
            pypi = jsonloads(response.text)

            grizzly_requirement_egg = grizzly_requirement

            # get grizzly version used in requirements.txt
            if re.match(r'^grizzly-loadtester(\[[^\]]*\])?$', grizzly_requirement):  # latest
                grizzly_version = pypi.get('info', {}).get('version', None)
            else:
                conditions: List[Callable[[versioning.Version], bool]] = []

                match = re.match(r'^(grizzly-loadtester(\[[^\]]*\])?)(.*?)$', grizzly_requirement)

                if match:
                    grizzly_requirement_egg = match.group(1)
                    condition_expression = match.group(3)

                    for condition in condition_expression.split(',', 1):
                        version_string = re.sub(r'^[^0-9]{1,2}', '', condition)
                        condition_version = versioning.parse(version_string)

                        if '>' in condition:
                            compare = condition_version.__le__ if '=' in condition else condition_version.__lt__
                        elif '<' in condition:
                            compare = condition_version.__ge__ if '=' in condition else condition_version.__gt__
                        else:
                            compare = condition_version.__eq__

                        conditions.append(compare)

                matched_version = None

                for available_version in pypi.get('releases', {}).keys():
                    try:
                        version = versioning.parse(available_version)
                        if len(conditions) > 0 and all([compare(version) for compare in conditions]):
                            matched_version = version
                    except versioning.InvalidVersion:
                        pass

                if matched_version is None:
                    print(f'!! could not resolve {grizzly_requirement} to one specific version available at pypi', file=sys.stderr)
                else:
                    grizzly_version = str(matched_version)

            if grizzly_version is not None:
                # get version from pypi, to be able to get locust version
                response = requests.get(
                    f'https://pypi.org/pypi/grizzly-loadtester/{grizzly_version}/json'
                )

                if response.status_code != 200:
                    print(f'!! unable to get grizzly {grizzly_version} package information from {response.url} ({response.status_code})', file=sys.stderr)
                else:
                    release_info = jsonloads(response.text)

                    for requires_dist in release_info.get('info', {}).get('requires_dist', []):
                        if not requires_dist.startswith('locust'):
                            continue

                        match = re.match(r'^locust \((.*?)\)$', requires_dist.strip())

                        if match:
                            locust_version = cast(str, match.group(1))
                        else:
                            locust_version = requires_dist.replace('locust', '').strip()

                        if locust_version is not None and locust_version.startswith('=='):
                            locust_version = locust_version[2:]

                        if len(locust_version or '') < 1:
                            print(f'!! unable to find locust version in "{requires_dist.strip()}" specified in pypi for grizzly-loadtester {grizzly_version}', file=sys.stderr)
                            locust_version = '(unknown)'
                        break

                    if locust_version is None:
                        print(f'!! could not find "locust" in requires_dist information for grizzly-loadtester {grizzly_version}', file=sys.stderr)

    if grizzly_version is None:
        grizzly_version = '(unknown)'
    else:
        match = re.match(r'^grizzly-loadtester\[([^\]]*)\]$', grizzly_requirement_egg)

        if match:
            grizzly_extras = [extra.strip() for extra in match.group(1).split(',')]
        else:
            grizzly_extras = []

    if locust_version is None:
        locust_version = '(unknown)'

    return (grizzly_version, grizzly_extras, ), locust_version


def list_images(args: Arguments) -> Dict[str, Any]:
    images: Dict[str, Any] = {}
    output = subprocess.check_output([
        f'{args.container_system}',
        'image',
        'ls',
        '--format',
        '{"name": "{{.Repository}}", "tag": "{{.Tag}}", "size": "{{.Size}}", "created": "{{.CreatedAt}}", "id": "{{.ID}}"}',
    ]).decode('utf-8')

    for line in output.split('\n'):
        if len(line) < 1:
            continue
        image = jsonloads(line)
        name = image['name']
        tag = image['tag']
        del image['name']
        del image['tag']

        version = {tag: image}

        if name not in images:
            images[name] = {}
        images[name].update(version)

    return images


def get_default_mtu(args: Arguments) -> Optional[str]:
    try:
        output = subprocess.check_output([
            f'{args.container_system}',
            'network',
            'inspect',
            'bridge',
            '--format',
            '{{ json .Options }}',
        ]).decode('utf-8')

        line, _ = output.split('\n', 1)
        network_options: Dict[str, str] = jsonloads(line)
        return network_options.get('com.docker.network.driver.mtu', '1500')
    except:
        return None


def requirements(execution_context: str) -> Callable[[Callable[..., int]], Callable[..., int]]:
    def wrapper(func: Callable[..., int]) -> Callable[..., int]:
        @wraps(func)
        def _wrapper(*args: Tuple[Any, ...], **kwargs: Dict[str, Any]) -> int:
            return func(*args, **kwargs)

        # a bit ugly, but needed for testability
        setattr(func, '__value__', execution_context)
        setattr(_wrapper, '__wrapped__', func)

        return _wrapper

    return wrapper


def get_distributed_system() -> Optional[str]:
    if which('docker') is not None:
        container_system = 'docker'
    elif which('podman') is not None:
        container_system = 'podman'
        print('!! podman might not work due to buildah missing support for `RUN --mount=type=ssh`: https://github.com/containers/buildah/issues/2835')
    else:
        print('neither "podman" nor "docker" found in PATH')
        return None

    rc, _ = subprocess.getstatusoutput(f'{container_system} compose version')

    if rc != 0:
        print(f'"{container_system} compose" not found in PATH')
        return None

    return container_system


def get_input(text: str) -> str:  # pragma: no cover
    return input(text).strip()


def ask_yes_no(question: str) -> None:
    answer = 'undefined'
    while answer.lower() not in ['y', 'n']:
        if answer != 'undefined':
            print('you must answer y (yes) or n (no)')
        answer = get_input(f'{question} [y/n]: ')

        if answer == 'n':
            raise KeyboardInterrupt()


def parse_feature_file(file: str) -> None:
    if len(grizzly_cli.SCENARIOS) > 0:
        return

    feature = feature_file_parser(file)

    grizzly_cli.FEATURE_DESCRIPTION = feature.name

    for scenario in feature.scenarios:
        grizzly_cli.SCENARIOS.append(scenario)


def find_metadata_notices(file: str) -> List[str]:
    with open(file) as fd:
        return [line.strip().replace('# grizzly-cli:notice ', '') for line in fd.readlines() if line.strip().startswith('# grizzly-cli:notice ')]


def find_variable_names_in_questions(file: str) -> List[str]:
    unique_variables: Set[str] = set()

    parse_feature_file(file)

    for scenario in grizzly_cli.SCENARIOS:
        for step in scenario.steps + scenario.background_steps or []:
            if not step.name.startswith('ask for value of variable'):
                continue

            match = re.match(r'ask for value of variable "([^"]*)"', step.name)

            if not match:
                raise ValueError(f'could not find variable name in "{step.name}"')

            unique_variables.add(match.group(1))

    return sorted(list(unique_variables))


def distribution_of_users_per_scenario(args: Arguments, environ: Dict[str, Any]) -> None:
    def _guess_datatype(value: str) -> Union[str, int, float, bool]:
        check_value = value.replace('.', '', 1)

        if check_value[0] == '-':
            check_value = check_value[1:]

        if check_value.isdecimal():
            if float(value) % 1 == 0:
                if value.startswith('0'):
                    return str(value)
                else:
                    return int(float(value))
            else:
                return float(value)
        elif value.lower() in ['true', 'false']:
            return value.lower() == 'true'
        else:
            return value

    class ScenarioProperties:
        name: str
        index: int
        identifier: str
        user: Optional[str]
        weight: float
        _iterations: Optional[int]
        _user_count: Optional[int]

        def __init__(
            self,
            name: str,
            index: int,
            weight: Optional[float] = None,
            user: Optional[str] = None,
            iterations: Optional[int] = None,
            user_count: Optional[int] = None,
        ) -> None:
            self.name = name
            self.index = index
            self.user = user
            self._iterations = iterations
            self.weight = weight or 1.0
            self.identifier = f'{index:03}'
            self._user_count = user_count

        @property
        def iterations(self) -> int:
            if self._iterations is None:  # pragma: no cover
                raise ValueError('iterations has not been set')

            return self._iterations

        @iterations.setter
        def iterations(self, value: int) -> None:
            self._iterations = value

        @property
        def user_count(self) -> int:
            if self._user_count is None:  # pragma: no cover
                raise ValueError('user count has not been set')
            return self._user_count

        @user_count.setter
        def user_count(self, value: int) -> None:
            self._user_count = value

        def is_fulfilled(self) -> bool:
            return self.user is not None and self._iterations is not None and self._user_count is not None

    distribution: Dict[str, ScenarioProperties] = {}
    variables = {key.replace('TESTDATA_VARIABLE_', ''): _guess_datatype(value) for key, value in environ.items() if key.startswith('TESTDATA_VARIABLE_')}

    def _pre_populate_scenario(scenario: Scenario, index: int) -> None:
        if scenario.name not in distribution:
            distribution[scenario.name] = ScenarioProperties(
                name=scenario.name,
                index=index,
                user=None,
                weight=None,
                iterations=None,
            )

    scenario_user_count_total: Optional[int] = None
    use_weights = True

    for index, scenario in enumerate(grizzly_cli.SCENARIOS):
        scenario_variables: Dict[str, Any] = {}
        if len(scenario.steps) < 1:
            raise ValueError(f'scenario "{scenario.name}" does not have any steps')

        _pre_populate_scenario(scenario, index=index + 1)

        if index == 0:  # background_steps is only processed for first scenario in grizzly
            for step in scenario.background_steps or []:
                if (step.name.endswith(' users') or step.name.endswith(' user')) and step.keyword == 'Given':
                    match = re.match(r'"([^"]*)" user(s)?', step.name)
                    if match:
                        scenario_user_count_total = int(round(float(Template(match.group(1)).render(**variables)), 0))

        if scenario_user_count_total is None:
            use_weights = False
            scenario_user_count_total = 0

        for step in (scenario.background_steps or []) + scenario.steps:
            if step.name.startswith('value for variable'):  # pragma: no cover
                match = re.match(r'value for variable "([^"]*)" is "([^"]*)"', step.name)
                if match:
                    try:
                        variable_name = match.group(1)
                        variable_value = Template(match.group(2)).render(**variables, **scenario_variables)
                        scenario_variables.update({variable_name: variable_value})
                    except:
                        continue
            elif step.name.startswith('a user of type'):
                match = re.match(r'a user of type "([^"]*)" (with weight "([^"]*)")?.*', step.name)
                if match:
                    distribution[scenario.name].user = match.group(1)
                    distribution[scenario.name].weight = int(float(Template(match.group(3) or '1.0').render(**variables, **scenario_variables)))
            elif step.name.startswith('repeat for'):
                match = re.match(r'repeat for "([^"]*)" iteration[s]?', step.name)
                if match:
                    distribution[scenario.name].iterations = int(round(float(Template(match.group(1)).render(**variables, **scenario_variables)), 0))
                    (distribution[scenario.name].iterations)
            elif any([pattern in step.name for pattern in ['users of type', 'user of type']]):
                match = re.match(r'"([^"]*)" user[s]? of type "([^"]*)".*', step.name)
                if match:
                    scenario_user_count = int(round(float(Template(match.group(1)).render(**variables, **scenario_variables)), 0))
                    scenario_user_count_total += scenario_user_count

                    distribution[scenario.name].user_count = scenario_user_count
                    distribution[scenario.name].user = match.group(2)

            if distribution[scenario.name].is_fulfilled():
                break

    scenario_count = len(distribution.keys())
    assert scenario_user_count_total is not None
    if scenario_count > scenario_user_count_total:
        raise ValueError(f'grizzly needs at least {scenario_count} users to run this feature')

    total_weight = 0
    total_iterations = 0
    for scenario in distribution.values():
        if scenario.user is None:
            raise ValueError(f'{scenario.name} does not have a user type')

        total_weight += scenario.weight
        total_iterations += scenario.iterations

    if use_weights:
        for scenario in distribution.values():
            scenario.user_count = ceil(scenario_user_count_total * (scenario.weight / total_weight))

        # smooth assigned user count based on weight, so that the sum of scenario.user_count == total_user_count
        total_user_count = sum([scenario.user_count for scenario in distribution.values()])
        user_overflow = total_user_count - scenario_user_count_total

        while user_overflow > 0:
            for scenario in dict(sorted(distribution.items(), key=lambda d: d[1].user_count, reverse=True)).values():
                if scenario.user_count <= 1:
                    continue

                scenario.user_count -= 1
                user_overflow -= 1

                if user_overflow < 1:
                    break

    def print_table_lines(max_length_iterations: int, max_length_users: int, max_length_description: int, max_length_errors: int) -> None:
        line = ['-' * 5, '-|-', '-' * 6, '|-', '-' * max_length_iterations, '|-', '-' * max_length_users, '|-', '-' * max_length_description, '-|']
        if not use_weights:
            line = line[:1] + line[3:]
            line[0] = f'{line[0]}-'

        if max_length_errors > 0:
            line += ['-' * (max_length_errors + 1), '-|']
        logger.info(''.join(line))

    rows: List[str] = []
    max_length_description = len('description')
    max_length_iterations = len('#iter')
    max_length_users = len('#user')
    max_length_errors = len('errors')

    message = f'\nfeature file {args.file} will execute in total {total_iterations} iterations divided on {len(grizzly_cli.SCENARIOS)} scenarios'
    if hasattr(args, 'environment_file') and args.environment_file is not None:
        message = f'{message} with environment file {environ["GRIZZLY_CONFIGURATION_FILE"]}'

    logger.info(f'{message}\n')

    errors: Dict[str, List[str]] = {}

    for scenario in distribution.values():
        # check for errors
        if scenario.user_count < 1:
            if scenario.name not in errors:
                errors.update({scenario.name: []})

            errors[scenario.name].append('no users assigned')

        if scenario.iterations < 1:
            if scenario.name not in errors:
                errors.update({scenario.name: []})

            errors[scenario.name].append('no iterations')

        # calculate max length out of all rows
        max_length_description = max(len(scenario.name), max_length_description)
        max_length_iterations = max(len(str(scenario.iterations)), max_length_iterations)
        max_length_users = max(len(str(scenario.user_count)), max_length_users)
        max_length_errors = max(len(', '.join(errors.get(scenario.name, []))), max_length_errors)

    # there was no errors, so reset max length for that column, so it won't be included
    if len(errors) < 1:
        max_length_errors = 0

    row_format = ['{:5} ', '  {:>6d}', '  {:>{}}', '  {:>{}}', '  {:<{}}']
    if not use_weights:
        row_format.pop(1)

    if len(errors) > 0:
        row_format.append('   {}')

    for scenario in distribution.values():
        row_format_args: List[Any] = [
            scenario.identifier,
            scenario.weight,
            scenario.iterations,
            max_length_iterations,
            scenario.user_count,
            max_length_users,
            scenario.name,
            max_length_description,
        ]

        # remove weights column
        if not use_weights:
            row_format_args.pop(1)

        if len(errors) > 0:
            row_format_args.append(
                ', '.join(errors.get(scenario.name, [])),
            )

        rows.append(''.join(row_format).format(*row_format_args))

    logger.info('each scenario will execute accordingly:\n')
    header_row_args: List[Any] = [
        'ident',
        'weight',
        '#iter', max_length_iterations,
        '#user', max_length_users,
        'description', max_length_description,
    ]

    header_row_format = ['{:5} ', '  {:>6}', '  {:>{}}', '  {:>{}}', '  {:<{}}']
    if not use_weights:
        header_row_format.pop(1)
        header_row_args.pop(1)

    if len(errors) > 0:
        header_row_format.append('   {}')
        header_row_args.append('errors')

    logger.info(''.join(header_row_format).format(*header_row_args))
    print_table_lines(max_length_iterations, max_length_users, max_length_description, max_length_errors)
    for row in rows:
        logger.info(row)
    print_table_lines(max_length_iterations, max_length_users, max_length_description, max_length_errors)

    if len(errors) > 0:
        arrow_width = len('ident') + 2 + max_length_iterations + 2 + max_length_users + 2 + max_length_description + 2
        if use_weights:
            arrow_width += len('weight') + 2
        message = f'''{" " * (1 + arrow_width)}^
+{"-" * arrow_width}+
|
+- there were errors when calculating user distribution and iterations per scenario, adjust user "weight", number of users or iterations per scenario
'''

        raise ValueError(message)
    else:
        logger.info('')

    for scenario in distribution.values():
        if scenario.iterations < scenario.user_count:
            raise ValueError(f'{scenario.name} will have {scenario.user_count} users to run {scenario.iterations} iterations, increase iterations or lower user count')

    if not args.yes:
        ask_yes_no('continue?')


def setup_logging(logfile: Optional[str] = None) -> None:
    logging_config: dict = {
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {
            'plain': {
                'format': '%(message)s',
            },
        },
        'handlers': {
            'console': {
                'class': 'logging.StreamHandler',
                'formatter': 'plain',
            },
        },
        'loggers': {
            'grizzly-cli': {
                'handlers': ['console'],
                'level': 'INFO',
                'propagate': False,
            },
        },
        'root': {
            'handlers': ['console'],
            'level': 'INFO',
        },
    }

    if logfile is not None:  # pragma: no cover
        logging_config['handlers']['file'] = {
            'class': 'logging.FileHandler',
            'filename': logfile,
            'formatter': 'plain',
        }

        logging_config['loggers']['grizzly-cli']['handlers'].append('file')
        logging_config['root']['handlers'].append('file')

    logging.config.dictConfig(logging_config)


def unflatten(key: str, value: Any) -> dict[str, Any]:
    paths: list[str] = key.split('.')

    # last node should have the value
    path = paths.pop()
    struct = {path: value}

    # build the struct from the inside out
    paths.reverse()

    for path in paths:
        struct = {path: {**struct}}

    return struct


def flatten(node: dict[str, Any], parents: Optional[list[str]] = None) -> dict[str, Any]:
    """Flatten a dictionary so each value key is the path down the nested dictionary structure."""
    flat: dict[str, Any] = {}
    if parents is None:
        parents = []

    for key, value in node.items():
        parents.append(key)
        if isinstance(value, dict):
            flat = {**flat, **flatten(value, parents)}
        else:
            flat['.'.join(parents)] = value

        parents.pop()

    return flat


def merge_dicts(merged: dict[str, Any], source: dict[str, Any]) -> dict[str, Any]:
    """Merge two dicts recursively, where `source` values takes precedance over `merged` values."""
    merged = deepcopy(merged)
    source = deepcopy(source)

    for key in source:
        if (
            key in merged
            and isinstance(merged[key], dict)
            and isinstance(source[key], Mapping)
        ):
            merged[key] = merge_dicts(merged[key], source[key])
        else:
            value = source[key]
            if isinstance(value, str) and value.lower() == 'none':  # pragma: no cover
                value = None
            merged[key] = value

    return merged


def chunker(value: str, size: int) -> list[str]:
    return list(
        map(lambda x: value[x * size:x * size + size],
            list(range(ceil(len(value) / size)))))


def get_indentation(file: Path) -> int:
    try:
        first_indent_line = file.read_text().splitlines()[1]
        return len(first_indent_line.rstrip()) - len(first_indent_line.strip())
    except IndexError:
        # use 2 as default indentation when it is not possible to detect from file
        return 2


class IndentDumper(Dumper):
    use_indent: ClassVar[int]

    @classmethod
    def use_indentation(cls, target: Path | int) -> type['IndentDumper']:
        cls.use_indent = get_indentation(target) if isinstance(target, Path) else target

        return cls

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

        self.best_indent = self.use_indent

    def increase_indent(self, flow: bool = False, indentless: bool = False) -> None:
        return super().increase_indent(flow, False)



import sys
import re

from typing import Any, Optional, IO, Sequence, cast
from argparse import ArgumentParser as CoreArgumentParser, Namespace, _SubParsersAction

from .markdown import MarkdownFormatter, MarkdownHelpAction
from .bashcompletion import BashCompletionAction, hook as bashcompletion_hook


ArgumentSubParser = _SubParsersAction


class ArgumentParser(CoreArgumentParser):
    def __init__(self, markdown_help: bool = False, bash_completion: bool = False, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

        self.markdown_help = markdown_help
        self.bash_completion = bash_completion

        if self.markdown_help:
            self.add_argument('--md-help', action=MarkdownHelpAction)

        if self.bash_completion:
            self.add_argument('--bash-completion', action=BashCompletionAction)

        self._optionals.title = 'optional arguments'

    def error_no_help(self, message: str) -> None:
        sys.stderr.write('{}: error: {}\n'.format(self.prog, message))
        sys.exit(2)

    def print_help(self, file: Optional[IO[str]] = None) -> None:
        '''Hook to make help more command line friendly, if there is markdown markers in the text.
        '''
        if not self.markdown_help:
            super().print_help(file)
            return

        if cast(type, self.formatter_class) is not MarkdownFormatter:
            original_description = self.description
            original_actions = self._actions

            # code block "markers" are not really nice to have in cli help
            if self.description is not None:
                self.description = '\n'.join([line for line in self.description.split('\n') if '```' not in line])
                self.description = self.description.replace('\n\n', '\n')

            for action in self._actions:
                if action.help is not None:
                    # remove any markdown link markers
                    action.help = re.sub(r'\[([^\]]*)\]\([^\)]*\)', r'\1', action.help)

        super().print_help(file)

        if cast(type, self.formatter_class) is not MarkdownFormatter:
            self.description = original_description
            self._actions = original_actions

    def parse_args(self, args: Optional[Sequence[str]] = None, namespace: Optional[Namespace] = None) -> Namespace:  # type: ignore
        """
        Hook to add `--bash-complete` to all parsers, if enabled for parser.
        """
        if self.bash_completion:
            bashcompletion_hook(self)

        return super().parse_args(args, namespace)



import os

from typing import Callable, List, Optional

from behave.model import Scenario

from .argparse import ArgumentSubParser
from .__version__ import __version__


EXECUTION_CONTEXT = os.getcwd()

STATIC_CONTEXT = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'static')

MOUNT_CONTEXT = os.environ.get('GRIZZLY_MOUNT_CONTEXT', EXECUTION_CONTEXT)

PROJECT_NAME = os.path.basename(EXECUTION_CONTEXT)

SCENARIOS: List[Scenario] = []

FEATURE_DESCRIPTION: Optional[str] = None


class register_parser:
    registered: List[Callable[[ArgumentSubParser], None]] = []
    order: Optional[int]

    def __init__(self, order: Optional[int] = None) -> None:
        self.order = order

    def __call__(self, func: Callable[[ArgumentSubParser], None]) -> Callable[[ArgumentSubParser], None]:
        if self.order is not None:
            self.registered.insert(self.order - 1, func)
        else:
            self.registered.append(func)

        return func


__all__ = [
    '__version__',
]



import argparse
import os
import sys

from shutil import which
from typing import Tuple, Optional, List
from traceback import format_exc

from .argparse import ArgumentParser
from .utils import ask_yes_no, get_distributed_system, get_dependency_versions, setup_logging
from .init import init
from .local import local
from .distributed import distributed
from .auth import auth
from .keyvault import keyvault
from . import __version__, register_parser


def _create_parser() -> ArgumentParser:
    parser = ArgumentParser(
        description=(
            'the command line interface for grizzly, which makes it easer to start a test with all features of grizzly wrapped up nicely.\n\n'
            'installing it is a matter of:\n\n'
            '```bash\n'
            'pip install grizzly-loadtester-cli\n'
            '```\n\n'
            'enable bash completion by adding the following to your shell profile:\n\n'
            '```bash\n'
            'eval "$(grizzly-cli --bash-completion)"\n'
            '```'
        ),
        markdown_help=True,
        bash_completion=True,
    )

    if parser.prog != 'grizzly-cli':
        parser.prog = 'grizzly-cli'

    parser.add_argument(
        '--version',
        nargs='?',
        default=None,
        const=True,
        choices=['all'],
        help='print version of command line interface, and exit. add argument `all` to get versions of dependencies',
    )

    sub_parser = parser.add_subparsers(dest='command')

    for create_parser in register_parser.registered:
        create_parser(sub_parser)

    return parser


def _parse_arguments() -> argparse.Namespace:
    parser = _create_parser()
    args = parser.parse_args()

    if hasattr(args, 'file'):
        # needed to support file names with spaces, which is escaped (sh-style)
        setattr(args, 'file', ' '.join(args.file))

    if args.version:
        if __version__ == '0.0.0':
            version = '(development)'
        else:
            version = __version__

        grizzly_versions: Optional[Tuple[Optional[str], Optional[List[str]]]] = None

        if args.version == 'all':
            grizzly_versions, locust_version = get_dependency_versions(False)
        else:
            grizzly_versions, locust_version = None, None

        print(f'grizzly-cli {version}')
        if grizzly_versions is not None:
            grizzly_version, grizzly_extras = grizzly_versions
            if grizzly_version is not None:
                print(f'└── grizzly {grizzly_version}', end='')
                if grizzly_extras is not None and len(grizzly_extras) > 0:
                    print(f' ── extras: {", ".join(grizzly_extras)}', end='')
                print('')

        if locust_version is not None:
            print(f'    └── locust {locust_version}')

        raise SystemExit(0)

    if args.command is None:
        parser.error('no command specified')

    if getattr(args, 'subcommand', None) is None and args.command not in ['init', 'auth']:
        parser.error_no_help(f'no subcommand for {args.command} specified')

    if args.command == 'dist':
        args.container_system = get_distributed_system()

        if args.container_system is None:
            parser.error_no_help('cannot run distributed')

        if args.registry is not None and not args.registry.endswith('/'):
            setattr(args, 'registry', f'{args.registry}/')
    elif args.command in ['init', 'auth']:
        setattr(args, 'subcommand', None)

    if args.subcommand == 'run':
        if args.command == 'dist':
            if args.limit_nofile < 10001 and not args.yes:
                print('!! this will cause warning messages from locust later on')
                ask_yes_no('are you sure you know what you are doing?')
        elif args.command == 'local':
            if which('behave') is None:
                parser.error_no_help('"behave" not found in PATH, needed when running local mode')

        if args.testdata_variable is not None:
            for variable in args.testdata_variable:
                try:
                    [name, value] = variable.split('=', 1)
                    os.environ[f'TESTDATA_VARIABLE_{name}'] = value
                except ValueError:
                    parser.error_no_help('-T/--testdata-variable needs to be in the format NAME=VALUE')

        if args.csv_prefix is None:
            if args.csv_interval is not None:
                parser.error_no_help('--csv-interval can only be used in combination with --csv-prefix')

            if args.csv_flush_interval is not None:
                parser.error_no_help('--csv-flush-interval can only be used in combination with --csv-prefix')
    elif args.command == 'dist' and args.subcommand == 'build':
        setattr(args, 'force_build', args.no_cache)
        setattr(args, 'build', not args.no_cache)

    log_file = getattr(args, 'log_file', None)
    setup_logging(log_file)

    return args


def _inject_additional_arguments_from_metadata(args: argparse.Namespace) -> argparse.Namespace:
    with open(args.file) as fd:
        file_metadata = [line.strip().replace('# grizzly-cli ', '').split(' ') for line in fd.readlines() if line.strip().startswith('# grizzly-cli ')]

    if len(file_metadata) < 1:
        return args

    argv = sys.argv[1:]
    for additional_arguments in file_metadata:
        try:
            if additional_arguments[0].strip().startswith('-'):
                raise ValueError()

            index = argv.index(additional_arguments[0]) + 1
            for zindex, additional_argument in enumerate(additional_arguments[1:]):
                argv.insert(index + zindex, additional_argument)
        except ValueError:
            print('?? ignoring {}'.format(' '.join(additional_arguments)))

    sys.argv = sys.argv[0:1] + argv

    return _parse_arguments()


def main() -> int:
    args: Optional[argparse.Namespace] = None

    try:
        args = _parse_arguments()

        if getattr(args, 'file', None) is not None and args.command not in ['keyvault']:
            args = _inject_additional_arguments_from_metadata(args)

        if args.command == 'local':
            rc = local(args)
        elif args.command == 'dist':
            rc = distributed(args)
        elif args.command == 'init':
            rc = init(args)
        elif args.command == 'auth':
            rc = auth(args)
        elif args.command == 'keyvault':
            rc = keyvault(args)
        else:
            raise ValueError(f'unknown command {args.command}')

        return rc
    except (KeyboardInterrupt, ValueError) as e:
        print('')
        if isinstance(e, ValueError):
            if args is not None and getattr(args, 'verbose', False):
                exception = format_exc()
            else:
                exception = str(e)

            print(exception)

        print('\n!! aborted grizzly-cli')
        return 1



import sys

from typing import Dict, Optional
from glob import glob
from os import getcwd
from os.path import sep as path_separator, exists, isfile
from fnmatch import filter as fnmatch_filter
from argparse import ArgumentTypeError


__all__ = [
    'BashCompletionTypes',
]

ESCAPE_CHARACTERS = {
    ' ': '\\ ',
    '(': '\\(',
    ')': '\\)',
}


class BashCompletionTypes:
    class File:
        _cwd: str = getcwd()

        def __init__(self, *args: str, missing_ok: bool = False) -> None:
            self.patterns = list(args)
            self.cwd = BashCompletionTypes.File._cwd
            self.missing_ok = missing_ok

        def __call__(self, value: str) -> str:
            if self.missing_ok:
                return value

            if not exists(value):
                raise ArgumentTypeError(f'{value} does not exist')

            if not isfile(value):
                raise ArgumentTypeError(f'{value} is not a file')

            matches = [match for pattern in self.patterns for match in fnmatch_filter([value], pattern)]

            if len(matches) < 1:
                raise ArgumentTypeError(f'{value} does not match {", ".join(self.patterns)}')

            return value

        def list_files(self, value: Optional[str]) -> Dict[str, str]:
            matches: Dict[str, str] = {}

            if value is not None:
                if value.endswith('\\') and sys.platform != 'win32':
                    value += ' '
                value = value.replace('\\ ', ' ').replace('\\(', '(').replace('\\)', ')')

            for pattern in self.patterns:
                for path in glob('**/{pattern}'.format(pattern=pattern), recursive=True):
                    path_match = path.replace('{cwd}{path_separator}'.format(cwd=self.cwd, path_separator=path_separator), '')

                    if path_match.startswith('.') or (value is not None and not path_match.startswith(value)):
                        continue

                    match: Optional[Dict[str, str]] = None

                    if path_separator in path_match:
                        try:
                            index_match = len(value or '')
                            index_sep = path_match[index_match:].index(path_separator) + index_match
                            match = {path_match[:index_sep].translate(str.maketrans(ESCAPE_CHARACTERS)): 'dir'}  # type: ignore
                        except ValueError:
                            pass

                    if match is None:
                        match = {path_match.translate(str.maketrans(ESCAPE_CHARACTERS)): 'file'}  # type: ignore

                    matches.update(match)

            return matches



import sys

from os import environ
from argparse import Namespace as Arguments
from typing import Optional
from pathlib import Path

from pyotp import TOTP

from . import register_parser
from .argparse import ArgumentSubParser


@register_parser()
def create_parser(sub_parser: ArgumentSubParser) -> None:
    # grizzly-cli auth
    auth_parser = sub_parser.add_parser('auth', description=(
        'grizzly stateless authenticator application'
    ))

    auth_parser.add_argument(
        'input',
        nargs='?',
        type=str,
        default=None,
        const=None,
        help=(
            'where to read OTP secret, nothing specified means environment variable OTP_SECRET, '
            '`-` means stdin and anything else is considered a file'
        )
    )

    if auth_parser.prog != 'grizzly-cli auth':  # pragma: no cover
        auth_parser.prog = 'grizzly-cli auth'


def auth(args: Arguments) -> int:
    secret: Optional[str] = None

    if args.input is None:
        secret = environ.get('OTP_SECRET', None)
        if secret is None:
            raise ValueError('environment variable OTP_SECRET is not set')
    elif args.input == '-':
        try:
            secret = sys.stdin.read().strip()
        except:
            pass
        finally:
            if secret is None or len(secret.strip()) < 1:
                raise ValueError('OTP secret could not be read from stdin')
    else:
        input_file = Path(args.input)

        if not input_file.exists():
            raise ValueError(f'file {input_file} does not exist')

        secret = input_file.read_text().strip()

        if ' ' in secret or len(secret.split('\n')) > 1 or secret == '':
            raise ValueError(f'file {input_file} does not seem to contain a single line with a valid OTP secret')

    try:
        totp = TOTP(secret)

        print(totp.now())
    except Exception as e:
        raise ValueError(f'unable to generate TOTP code: {e}')

    return 0



#!/usr/bin/env python

import sys

from typing import List
from os import path
from json import loads as jsonloads
from io import StringIO

import requests

from piplicenses import CustomNamespace, FormatArg, FromArg, OrderArg, create_output_string
from pytablewriter import MarkdownTableWriter

URL_MAP = {}

REPO_ROOT = path.realpath(path.join(path.dirname(__file__), '..'))


def generate_license_table() -> List[str]:
    args = CustomNamespace()
    args.format_ = FormatArg.JSON
    args.from_ = FromArg.MIXED
    args.order = OrderArg.LICENSE
    args.summary = False
    args.with_authors = False
    args.with_urls = True
    args.with_description = False
    args.with_license_file = True
    args.no_license_path = False
    args.with_license_file = False
    args.ignore_packages = []
    args.packages = []
    args.fail_on = None
    args.allow_only = None
    args.with_system = False
    args.filter_strings = False
    args.with_maintainers = False
    args.no_version = False
    args.python = sys.executable

    licenses = jsonloads(create_output_string(args))
    headers = ['Name', 'Version', 'License']

    table_contents: List[List[str]] = []

    for license in licenses:
        name = license['Name']
        # https://stackoverflow.com/questions/39577984/what-is-pkg-resources-0-0-0-in-output-of-pip-freeze-command
        if name.startswith('grizzly-') or name in ['pkg-resources']:
            continue

        if license['URL'] == 'UNKNOWN':
            try:
                response = requests.get(f'https://pypi.org/pypi/{name}/json')

                if response.status_code != 200:
                    raise ValueError(f'{response.url} returned {response.status_code}')

                result = jsonloads(response.text)

                info = result.get('info', None) or {}
                project_urls = info.get('project_urls', None) or {}

                url = project_urls.get(
                    'Homepage',
                    project_urls.get(
                        'Home',
                        info.get(
                            'project_url',
                            info.get(
                                'package_url',
                                URL_MAP.get(
                                    name, None,
                                ),
                            ),
                        ),
                    ),
                )

                if url is None:
                    raise ValueError(f'no URL found on {response.url} or in static map')

                license['URL'] = url
            except Exception as e:
                print(f'!! you need to find an url for package "{name}": {str(e)}', file=sys.stderr)
                sys.exit(1)

        name = f'[{name}]({license["URL"]})'

        table_contents.append([
            name,
            license['Version'],
            license['License'],
        ])

    writer = MarkdownTableWriter(
        headers=headers,
        value_matrix=table_contents,
        margin=1,
    )

    writer.stream = StringIO()
    writer.write_table()

    license_table = ['### Python dependencies\n'] + [f'{row}\n' for row in writer.stream.getvalue().strip().split('\n')]

    return license_table


def main() -> int:
    with open(path.join(REPO_ROOT, 'LICENSE.md')) as fd:
        contents = fd.readlines()

    license_table = generate_license_table()
    contents[0] = f'#{contents[0]}'
    license_contents = contents + ['\n', '## Third party licenses\n', '\n'] + license_table[:-1]

    print(''.join(license_contents))

    return 0


if __name__ == '__main__':
    sys.exit(main())



from __future__ import annotations

import re
from pathlib import Path
from contextlib import suppress
from textwrap import dedent
from typing import Any, Iterable, ClassVar, cast
from base64 import b64decode

import yaml
from azure.identity import AzureCliCredential, ManagedIdentityCredential, ChainedTokenCredential
from azure.keyvault.secrets import SecretClient, KeyVaultSecret
from jinja2 import Environment
from jinja2.lexer import Token, TokenStream
from jinja2_simple_tags import StandaloneTag
from behave.parser import parse_feature
from behave.model import Scenario

from grizzly_cli.utils import IndentDumper, merge_dicts, logger, unflatten


def get_context_root() -> Path:
    possible_context_root = Path.cwd().rglob('environment.py')
    try:
        return next(possible_context_root).parent
    except StopIteration:
        raise ValueError('context root not found, are you in a grizzly project?')


class ScenarioTag(StandaloneTag):
    tags = {'scenario'}

    def preprocess(
        self, source: str, name: str | None, filename: str | None = None
    ) -> str:
        self._source = source

        return cast(str, super().preprocess(source, name, filename))

    @classmethod
    def get_scenario_text(cls, name: str, file: Path) -> str:
        content = file.read_text()

        content_skel = re.sub(r'\{%.*%\}', '', content)
        content_skel = re.sub(r'\{\$.*\$\}', '', content_skel)

        assert len(content.splitlines()) == len(content_skel.splitlines()), 'oops, there is not a 1:1 match between lines!'

        feature = parse_feature(content_skel, filename=file.as_posix())
        scenarios = cast(list[Scenario], feature.scenarios)
        lines = content.splitlines()

        for scenario_index, scenario in enumerate(scenarios):
            if scenario.name == name:
                break

        # check if there are scenarios after our scenario in the source
        next_scenario: Scenario | None = None
        with suppress(IndexError):
            next_scenario = scenarios[scenario_index + 1]

        if next_scenario is None:  # last scenario, take everything until the end
            scenario_lines = lines[scenario.line:]
        else:  # take everything up until where the next scenario starts
            scenario_lines = lines[scenario.line:next_scenario.line - 1]
            if scenario_lines[-1] == '':  # if last line is an empty line, lets remove it
                scenario_lines.pop()

        # remove any scenario text/comments
        if scenario_lines[0].strip() == '"""':
            try:
                offset = scenario_lines[1:].index(scenario_lines[0]) + 1 + 1
            except:
                offset = 0

            scenario_lines = scenario_lines[offset:]

        # first line can have incorrect indentation
        scenario_lines[0] = dedent(scenario_lines[0])

        return '\n'.join(scenario_lines)

    def render(self, scenario: str, feature: str, **variables: str) -> str:
        feature_file = Path(feature)

        # check if relative to parent feature file
        if not feature_file.exists():
            feature_file = (self.environment.feature_file.parent / feature).resolve()

        scenario_content = self.get_scenario_text(scenario, feature_file)

        ignore_errors = getattr(self.environment, 'ignore_errors', False)

        # <!-- sub-render included scenario
        errors_unused: set[str] = set()
        errors_undeclared: set[str] = set()

        # tag has specified variables, so lets "render"
        for name, value in variables.items():
            variable_template = f'{{$ {name} $}}'
            if variable_template not in scenario_content:
                errors_unused.add(name)
                continue

            scenario_content = scenario_content.replace(variable_template, str(value))

        # look for sub-variables that has not been rendered
        if not ignore_errors:
            if '{$' in scenario_content and '$}' in scenario_content:
                matches = re.finditer(r'\{\$ ([^$]+) \$\}', scenario_content, re.MULTILINE)

                for match in matches:
                    errors_undeclared.add(match.group(1))

            if len(errors_undeclared) + len(errors_unused) > 0:
                scenario_identifier = f'{feature}#{scenario}'
                buffer_error: list[str] = []
                if len(errors_unused) > 0:
                    errors_unused_message = "\n  ".join(errors_unused)
                    buffer_error.append(f'the following variables has been declared in scenario tag but not used in {scenario_identifier}:\n  {errors_unused_message}')
                    buffer_error.append('')

                if len(errors_undeclared) > 0:
                    errors_undeclared_message = "\n  ".join(errors_undeclared)
                    buffer_error.append(f'the following variables was used in {scenario_identifier} but was not declared in scenario tag:\n  {errors_undeclared_message}')
                    buffer_error.append('')

                message = '\n'.join(buffer_error)
                raise ValueError(message)

        # check if we have nested statements (`{% .. %}`), and render again if that is the case
        if '{%' in scenario_content and '%}' in scenario_content:
            environment = self.environment.overlay()
            environment.feature_file = feature_file
            template = environment.from_string(scenario_content)
            scenario_content = template.render()
        # // -->

        return scenario_content

    def filter_stream(self, stream: TokenStream) -> TokenStream | Iterable[Token]:  # type: ignore[return]
        """Everything outside of `{% scenario ... %}` (and `{% if ... %}...{% endif %}`) should be treated as "data", e.g. plain text.

        Overloaded from `StandaloneTag`, must match method signature, which is not `Generator`, even though we yield
        the result instead of returning.
        """
        in_scenario = False
        in_block_comment = False
        in_condition = False
        in_variable = False

        variable_begin_pos = -1
        variable_end_pos = 0
        block_begin_pos = -1
        block_end_pos = 0
        source_lines = self._source.splitlines()

        for token in stream:
            if token.type == 'block_begin':
                if stream.current.value in self.tags:  # {% scenario ... %}
                    in_scenario = True
                    current_line = source_lines[token.lineno - 1].lstrip()
                    in_block_comment = current_line.startswith('#')
                    block_begin_pos = self._source.index(token.value, block_begin_pos + 1)
                elif stream.current.value in ['if', 'endif']:  # {% if <condition> %}, {% endif %}
                    in_condition = True

            if in_scenario:
                if token.type == 'block_end' and in_block_comment:
                    in_block_comment = False
                    block_end_pos = self._source.index(token.value, block_begin_pos)
                    token_value = self._source[block_begin_pos:block_end_pos + len(token.value)]
                    filtered_token = Token(token.lineno, 'data', token_value)
                elif in_block_comment:
                    continue
                else:
                    filtered_token = token
            elif in_condition:
                filtered_token = token
            else:
                if token.type == 'variable_begin':
                    # Find variable start in the source
                    variable_begin_pos = self._source.index(token.value, variable_begin_pos + 1)
                    in_variable = True
                    continue
                elif token.type == 'variable_end':
                    # Find variable end in the source
                    variable_end_pos = self._source.index(token.value, variable_begin_pos)
                    # Extract the variable definition substring and use as token value
                    token_value = self._source[variable_begin_pos:variable_end_pos + len(token.value)]
                    in_variable = False
                elif in_variable:  # Variable templates is yielded when the whole block has been processed
                    continue
                else:
                    token_value = token.value

                filtered_token = Token(token.lineno, 'data', token_value)

            yield filtered_token

            if token.type == 'block_end':
                if in_scenario:
                    in_scenario = False

                if in_condition:
                    in_condition = False


class MergeYamlTag(StandaloneTag):  # pragma: no cover
    tags: ClassVar[set[str]] = {'merge'}

    def preprocess(
        self, source: str, name: str | None, filename: str | None = None,
    ) -> str:
        self._source = source
        return cast(str, super().preprocess(source, name, filename))

    def render(self, filename: str, *filenames: str) -> str:
        buffer: list[str] = []

        files = [filename, *filenames]

        for file in files:
            merge_file = Path(file)

            # check if relative to parent feature file
            if not merge_file.exists():
                merge_file = (self.environment.source_file.parent / merge_file).resolve()

            if not merge_file.exists():
                raise FileNotFoundError(merge_file)

            merge_content = merge_file.read_text()

            if merge_content[0:3] != '---':
                buffer.append('---')

            buffer.append(merge_content)

        if self._source[0:3] != '---':
            buffer.append('---')

        return '\n'.join(buffer)

    def filter_stream(self, stream: TokenStream) -> TokenStream | Iterable[Token]:  # type: ignore[return]
        """Everything outside of `{% merge ... %}` should be treated as "data", e.g. plain text."""
        in_merge = False
        in_variable = False
        in_block_comment = False

        variable_begin_pos = -1
        variable_end_pos = 0
        block_begin_pos = -1
        block_end_pos = 0
        source_lines = self._source.splitlines()

        for token in stream:
            if token.type == 'block_begin' and stream.current.value in self.tags:
                in_merge = True
                current_line = source_lines[token.lineno - 1].lstrip()
                in_block_comment = current_line.startswith('#')
                block_begin_pos = self._source.index(token.value, block_begin_pos + 1)

            if not in_merge:
                if token.type == 'variable_end':
                    # Find variable end in the source
                    variable_end_pos = self._source.index(token.value, variable_begin_pos)
                    # Extract the variable definition substring and use as token value
                    token_value = self._source[variable_begin_pos:variable_end_pos + len(token.value)]
                    in_variable = False
                elif token.type == 'variable_begin':
                    # Find variable start in the source
                    variable_begin_pos = self._source.index(token.value, variable_begin_pos + 1)
                    in_variable = True
                else:
                    token_value = token.value

                if in_variable:
                    # While handling in-variable tokens, withhold values until
                    # the end of the variable is reached
                    continue

                filtered_token = Token(token.lineno, 'data', token_value)
            elif token.type == 'block_end' and in_block_comment:
                in_block_comment = False
                block_end_pos = self._source.index(token.value, block_begin_pos)
                token_value = self._source[block_begin_pos:block_end_pos + len(token.value)]
                filtered_token = Token(token.lineno, 'data', token_value)
            elif in_block_comment:
                continue
            else:
                filtered_token = token

            yield filtered_token

            if in_merge and token.type == 'block_end':
                in_merge = False


def get_keyvault_client(url: str) -> SecretClient:
    credential = ChainedTokenCredential(ManagedIdentityCredential(), AzureCliCredential())
    return SecretClient(vault_url=url, credential=credential)


def _get_metadata(content_type: str, name: str) -> str | None:
    try:
        start = content_type.index(f'{name}:') + len(name) + 1
    except ValueError:
        return None

    try:
        end = content_type.index(',', start)
    except ValueError:
        end = len(content_type)

    try:
        value = content_type[start:end]
    except IndexError:
        value = None

    return value


def _write_file(root: Path, content_type: str, encoded_content: str) -> str:
    file_name = _get_metadata(content_type, 'file')

    if file_name is None:
        raise ValueError('could not find `file:` in content type')

    file = root / 'files' / file_name
    file.parent.mkdir(parents=True, exist_ok=True)
    complete = True

    if 'chunk' in content_type:  # 0 = chunk:N, 1 = chunks:T
        chunk_index = int(_get_metadata(content_type, 'chunk') or 0)
        chunk_count = int(_get_metadata(content_type, 'chunks') or 1)
        content_buffer: list[str] = []

        logger.debug('%% processing chunk %d of %d for %s', chunk_index + 1, chunk_count, file_name)

        if chunk_index == 0 and file.exists():
            file.unlink()

        try:
            content_buffer.append(file.read_text())
        except FileNotFoundError:
            if chunk_index > 0:  # pragma: no cover
                raise

        content_buffer.append(encoded_content)

        if chunk_index < chunk_count - 1:
            file.write_text(''.join(content_buffer))
            complete = False
        else:
            logger.debug('%% completed %s with chunk %d of %d', file_name, chunk_index + 1, chunk_count)
            encoded_content = ''.join(content_buffer)
            complete = True

    relative_file = file.as_posix().replace(root.as_posix(), '')[1:]

    if complete:
        content = b64decode(encoded_content)
        file.write_bytes(content)
        logger.debug('%% wrote %s (size %d)', relative_file, len(content))

    return relative_file


def _import_files(client: SecretClient, root: Path, secret: KeyVaultSecret) -> str:
    if secret.value is None:
        message = f'secret {secret.name} has no value'
        raise ValueError(message)

    file_keys = secret.value.split(',')

    for file_key in file_keys:
        file_secret = client.get_secret(file_key)
        if file_secret.properties.content_type is None:
            message = f'secret {file_key} has no content type'
            raise ValueError(message)

        if file_secret.value is None:
            message = f'secret {file_key} has no value'
            raise ValueError(message)

        logger.debug('%% keyvault key %s', file_key)

        # nested files
        if file_secret.properties.content_type.startswith('files'):
            conf_value = _import_files(client, root, file_secret)
            logger.debug('%% nested files configuration value is %s', conf_value)
        else:
            conf_value = _write_file(root, file_secret.properties.content_type, file_secret.value)

    return conf_value


def load_configuration(configuration_file: str) -> str:
    file = Path(configuration_file)

    if not file.exists():
        message = f'{file.as_posix()} does not exist'
        raise ValueError(message)

    if file.suffix not in ['.yml', '.yaml']:
        message = 'configuration file must have file extension yml or yaml'
        raise ValueError(message)

    configuration = load_configuration_file(file)

    load_from_keyvault = ((configuration or {}).get('configuration', None) or {}).get('keyvault', None)

    if load_from_keyvault is not None:
        client = get_keyvault_client(load_from_keyvault)
        context_root = get_context_root()
        environment = configuration.get('configuration', {}).get('env', file.stem)

        loaded_keyvault_configuration, number_of_keyvault_secrets = load_configuration_keyvault(client, environment, context_root)
        keyvault_configuration = {'configuration': loaded_keyvault_configuration}

        configuration = merge_dicts(keyvault_configuration, configuration)

        logger.info('loaded %d secrets from keyvault %s', number_of_keyvault_secrets, load_from_keyvault)

    environment_lock_file = file.parent / f'{file.stem}.lock{file.suffix}'

    with environment_lock_file.open('w') as fd:
        yaml.dump(configuration, fd, Dumper=IndentDumper.use_indentation(file), default_flow_style=False, sort_keys=False, allow_unicode=True)

    return configuration_file.replace(file.name, f'{file.stem}.lock{file.suffix}')


def load_configuration_file(file: Path) -> dict[str, Any]:
    """Load a grizzly environment file and flatten the structure."""
    configuration: dict[str, Any] = {}

    environment = Environment(autoescape=False, extensions=[MergeYamlTag])
    environment.extend(source_file=file)
    loader = yaml.SafeLoader

    yaml_template = environment.from_string(file.read_text())
    yaml_content = yaml_template.render()

    yaml_configurations = list(yaml.load_all(yaml_content, Loader=loader))
    yaml_configurations.reverse()
    for yaml_configuration in yaml_configurations:
        configuration = merge_dicts(configuration, yaml_configuration)

    return configuration


def load_configuration_keyvault(client: SecretClient, environment: str, root: Path) -> tuple[dict[str, Any], int]:
    environment_filter = ['global', environment]

    secret_properties = client.list_properties_of_secrets()

    keys: dict[str, str] = {}
    configuration: dict[str, Any] = {}

    # loop through all secrets to find the ones that match the environment filter
    for secret_property in secret_properties:
        if (
            secret_property.name is None
            or not secret_property.name.startswith('grizzly--')
            or 'noconf' in (secret_property.content_type or '')  # skip chunked secrets, we'll get them later
        ):
            continue

        _, target_environment, name = secret_property.name.split('--', 2)

        # hm, chunk suffix but no chunk in name?
        if '--' in name:
            logger.debug(f'secret {secret_property.name} with name {name} contains `--`, skipping')
            continue

        name = name.replace('-', '.')

        if target_environment not in environment_filter:
            continue

        keys.update({secret_property.name: name})

    imported_secrets = 0

    # get the actual value for all secrets that matched environment filter
    for secret_key, conf_key in keys.items():
        secret = client.get_secret(secret_key)

        content_type = secret.properties.content_type

        if secret.value is None:
            logger.error(f'! secret {secret_key} has no value, skipping')
            continue

        conf_value = secret.value

        no_conf = False

        if content_type is not None:
            no_conf = 'noconf' in content_type

            if content_type.startswith('files'):
                conf_value = _import_files(client, root, secret)
                if all(keyword in secret_key for keyword in ['mq', 'key']):
                    conf_value = Path(conf_value).with_suffix('').as_posix()
            elif content_type.startswith('file:'):
                conf_value = _write_file(root, content_type, secret.value)
            else:
                message = f'unknown content type for secret {secret_key}: {content_type}'
                raise ValueError(message)

        if not no_conf:
            logger.debug('mapping %s to %s', conf_key, conf_value if content_type is not None else '******')
            configuration_branch = unflatten(conf_key, conf_value)
            configuration = merge_dicts(configuration_branch, configuration)
            imported_secrets += 1

    return configuration, imported_secrets



from __future__ import annotations

import logging
import re

from typing import Any
from pathlib import Path
from argparse import Namespace as Arguments, ArgumentParser as CoreArgumentParser
from base64 import b64encode
from dataclasses import dataclass
from contextlib import suppress

import yaml
from azure.core.exceptions import ResourceNotFoundError, ClientAuthenticationError, ServiceRequestError
from azure.keyvault.secrets import SecretClient

from . import register_parser
from .argparse import ArgumentSubParser
from .argparse.bashcompletion import BashCompletionTypes
from .utils import flatten, unflatten, IndentDumper, merge_dicts, logger, chunker
from .utils.configuration import get_context_root, load_configuration_file, load_configuration_keyvault, get_keyvault_client

# disable azure.identity warning logs if authentication fails
azure_logger = logging.getLogger('azure')
azure_logger.setLevel(logging.ERROR)


KEYWORDS = ['password', 'secret', 'key', 'cert', 'token']
COMMON_FALSE_POSITIVES = ['RootManageSharedAccessKey']
KEYVAULT_MAX_SIZE = 25600 - 100  # 100 bytes buffer


@dataclass
class KeyvaultSecretHolder:
    """
    Keyvault secret holder.
    """
    name: str
    content_type: str | None
    value: str


@register_parser()
def create_parser(sub_parser: ArgumentSubParser) -> None:
    # grizzly-cli keyvault
    keyvault_parser = sub_parser.add_parser('keyvault', description=(
        'grizzly keyvault integration'
    ))

    keyvault_parser.add_argument(
        '-f', '--file',
        dest='env_file',
        type=BashCompletionTypes.File('*.yaml', '*.yml', missing_ok=True),
        required=True,
        help='path to the grizzly environment configuration file',
    )

    keyvault_parser.add_argument(
        '--vault-name',
        dest='keyvault',
        metavar='URL',
        type=str,
        help='keyvault url',
        default=None,
        required=False,
    )

    if keyvault_parser.prog != 'grizzly-cli keyvault':
        keyvault_parser.prog = 'grizzly-cli keyvault'

    sub_parser = keyvault_parser.add_subparsers(dest='subcommand')

    create_import_parser(sub_parser)
    create_export_parser(sub_parser)
    create_diff_parser(sub_parser)


def add_generic_arguments(parser: CoreArgumentParser) -> None:
    parser.add_argument(
        '--verbose',
        action='store_true',
        help='enable verbose output',
    )

    parser.add_argument(
        '--dry-run',
        action='store_true',
        help='do not write to keyvault',
    )


def create_diff_parser(sub_parser: ArgumentSubParser) -> None:
    # grizzly-cli keyvault diff
    diff_parser = sub_parser.add_parser('diff', description='compare two environment configuration files')

    diff_parser.add_argument(
        'orig_file',
        nargs=None,
        type=BashCompletionTypes.File('*.yaml', '*.yml'),
        help='file to compare -f/--file with')

    if diff_parser.prog != 'grizzly-cli keyvault diff':
        diff_parser.prog = 'grizzly-cli keyvault diff'


def create_import_parser(sub_parser: ArgumentSubParser) -> None:
    # grizzly-cli keyvault import
    import_parser = sub_parser.add_parser('import', description=(
        'import grizzly environment configuration from keyvault secrets'
    ))

    add_generic_arguments(import_parser)

    if import_parser.prog != 'grizzly-cli keyvault import':
        import_parser.prog = 'grizzly-cli keyvault import'


def create_export_parser(sub_parser: ArgumentSubParser) -> None:
    # grizzly-cli keyvault export
    export_parser = sub_parser.add_parser('export', description=(
        'export grizzly environment configuration to keyvault secrets'
    ))

    export_parser.add_argument(
        '-g', '--global',
        dest='global_configuration',
        metavar='KEY',
        action='append',
        default=[],
        type=str,
        required=False,
        help='environment configuration keys that are global',
    )

    add_generic_arguments(export_parser)

    if export_parser.prog != 'grizzly-cli keyvault export':
        export_parser.prog = 'grizzly-cli keyvault export'


def _keyvault_normalize(value: str) -> str:
    return re.sub(r'[^a-zA-Z0-9]', '-', value)


def encode_mq_certificate(root: Path, environment: str, key: str, base_cert_name: str) -> list[KeyvaultSecretHolder]:
    certificate_files = root.glob(f'{base_cert_name}.*')
    secrets: list[KeyvaultSecretHolder] = []

    references: list[str] = []

    for certificate_file in certificate_files:
        key_name = f'grizzly--{environment}--{_keyvault_normalize(certificate_file.name)}'
        references.append(key_name)
        secrets.extend(encode_file(key_name, certificate_file, no_conf=True))

    secrets.append(
        KeyvaultSecretHolder(
            name=key,
            content_type='files',
            value=','.join(references),
        )
    )

    return secrets


def encode_file(key: str, file: str | Path, *, no_conf: bool = False) -> list[KeyvaultSecretHolder]:
    secrets: list[KeyvaultSecretHolder] = []
    if isinstance(file, str):
        file = Path(file)

    value = b64encode(file.read_bytes()).decode('utf-8')

    content_type_extra = ''

    if no_conf:
        content_type_extra = ',noconf'

    if len(value) > KEYVAULT_MAX_SIZE:
        references: list[str] = []
        chunks = chunker(value, KEYVAULT_MAX_SIZE)
        for index, chunk in enumerate(chunks):
            chunk_key = f'{key}--{index}'
            references.append(chunk_key)

            logger.debug(f'% creating chunk {index + 1} of {len(chunks)} in key {chunk_key}')

            secrets.append(KeyvaultSecretHolder(
                name=chunk_key,
                content_type=f'file:{file.name},chunk:{index},chunks:{len(chunks)}{content_type_extra}',
                value=chunk,
            ))

        secrets.append(KeyvaultSecretHolder(
            name=key,
            content_type=f'files{content_type_extra}',
            value=','.join(references),
        ))
    else:
        secrets.append(KeyvaultSecretHolder(
            name=key,
            content_type=f'file:{file.name}{content_type_extra}',
            value=value,
        ))

    return secrets


def _should_export(key: str, secret: Any) -> bool:
    should: list[bool] = []

    if key == 'keyvault' and secret.endswith('.vault.azure.net'):
        return False

    for keyword in KEYWORDS:
        key_should = keyword in key.lower() and not f'.{keyword}.' in key
        secret_should = isinstance(secret, str) and keyword in secret.lower() and secret not in COMMON_FALSE_POSITIVES

        # if key should be exported, but not the secret and the secret is not a string... it shouldn't
        if key_should and not secret_should:
            key_should = isinstance(secret, str)

        should.append(key_should or secret_should)

    return any(should)


def _determine_environment(global_configuration: list[str], environment: str, key: str) -> str:
    return 'global' if any(key.startswith(keyword) for keyword in global_configuration) else environment


def _build_key_name(environment: str, key: str) -> str:
    return f'grizzly--{environment}--{_keyvault_normalize(key)}'


def diff(left_file_name: str, right_file_name: str) -> int:
    left_config_file = Path(left_file_name)
    right_config_file = Path(right_file_name)

    if not left_config_file.exists():
        raise ValueError(f'environment configuration file {left_config_file.as_posix()} does not exist')

    if not right_config_file.exists():
        raise ValueError(f'environment configuration file {right_config_file.as_posix()} does not exist')

    left_config = flatten(load_configuration_file(right_config_file)['configuration'])
    right_config = flatten(load_configuration_file(left_config_file)['configuration'])

    logger.info(f'- {right_config_file.as_posix()}')
    logger.info(f'+ {left_config_file.as_posix()}')
    logger.info('')

    diffed_keys: set[str] = set()

    for key, value in left_config.items():
        if key not in right_config or right_config.get(key, None) != value:
            logger.error(f'- {key}: {value} != {right_config.get(key, None)}')
            diffed_keys.add(key)

    for key, value in right_config.items():
        if key not in left_config or left_config.get(key, None) != value and key not in diffed_keys:
            logger.error(f'+ {key}: {left_config.get(key, None)} != {value}')

    return 0


def keyvault_import(client: SecretClient, environment: str, args: Arguments, root: Path, configuration: dict[str, Any]) -> int:
    # unflatten existing configuration
    configuration_unflatten: dict[str, Any] = {}

    for conf_key, conf_value in configuration.items():
        configuration_branch = unflatten(conf_key, conf_value)
        configuration_unflatten = merge_dicts(configuration_branch, configuration_unflatten)

    configuration = configuration_unflatten

    keyvault_configuration, imported_secrets = load_configuration_keyvault(client, environment, root)

    configuration = merge_dicts(keyvault_configuration, configuration)

    env_file = Path(args.env_file)
    _dict_to_yaml(env_file, {'configuration': configuration}, indentation=env_file)

    logger.info('\nimported %d secrets from %s to %s', imported_secrets, client.vault_url, env_file.as_posix())

    return 0


def keyvault_export(client: SecretClient, environment: str, args: Arguments, root: Path, configuration: dict[str, Any]) -> int:
    """
    From grizzly to keyvault.
    """
    secrets: list[KeyvaultSecretHolder] = []

    environment_file = Path(args.env_file)

    safe_configuration = {}

    for key, secret in configuration.items():
        if not _should_export(key, secret):
            safe_configuration.update({key: secret})
            continue

        key_environment = _determine_environment(args.global_configuration, environment, key)
        key_name = _build_key_name(key_environment, key)

        if 'file' in key:
            if 'mq' in key:
                secrets.extend(encode_mq_certificate(root, key_environment, key_name, secret))
            else:
                secrets.extend(encode_file(key_name, secret))
        else:
            secrets.append(KeyvaultSecretHolder(
                name=key_name,
                content_type=None,
                value=secret,
            ))

    created_secrets_count = 0

    for secret in secrets:
        # check if secret already exists
        try:
            current_value = client.get_secret(secret.name)
            if current_value.value != secret.value or current_value.properties.content_type != secret.content_type:
                raise ResourceNotFoundError

            # it exists with the same value, skip
            continue
        except ResourceNotFoundError:  # great, import it into the keyvault
            pass

        logger.debug(f'% keyvault secret {secret.name} with content type {secret.content_type}')

        if not args.dry_run:
            client.set_secret(secret.name, secret.value, content_type=secret.content_type)

        created_secrets_count += 1

    already_exists = len(secrets) - created_secrets_count

    if not args.dry_run:
        unsafe_environment_file = environment_file.rename(environment_file.with_suffix(f'.unsafe{environment_file.suffix}'))
        safe_yaml_configuration: dict[str, Any] = {}

        for key, value in safe_configuration.items():
            safe_yaml_configuration = merge_dicts(safe_yaml_configuration, unflatten(key, value))

        if safe_yaml_configuration.get('keyvault', None) is None:
            safe_yaml_configuration.update({'keyvault': client.vault_url})

        safe_yaml_configuration = {'configuration': safe_yaml_configuration}

        # remove contents in safe file, before writing the safe configuration
        _dict_to_yaml(environment_file, safe_yaml_configuration, indentation=unsafe_environment_file)
    else:
        unsafe_environment_file = environment_file.with_suffix(f'.unsafe{environment_file.suffix}')

    logger.info(
        f'created {created_secrets_count} ({already_exists} already existed) secrets in keyvault {client.vault_url} '
        f'and saved the safe environment configuration in {environment_file.as_posix()}'
    )
    logger.warning(f'! the unsafe environment configuration is still present in {unsafe_environment_file.as_posix()}')

    return 0


def _dict_to_yaml(file: Path, content: dict[str, Any], *, indentation: Path | int) -> None:
    file.write_text('')

    with file.open('w') as fd:
        yaml.dump(content, fd, Dumper=IndentDumper.use_indentation(indentation), default_flow_style=False, sort_keys=False, allow_unicode=True)


def _extract_metadata(env_file: str) -> tuple[str, str | None, dict[str, Any]]:
    file = Path(env_file)
    configuration = load_configuration_file(file).get('configuration', {})

    keyvault = configuration.get('keyvault', None)

    return (configuration.get('env', None) or file.stem, keyvault, flatten(configuration))


def keyvault(args: Arguments) -> int:
    grizzly_context_root = get_context_root()

    with suppress(AttributeError):
        if args.verbose or args.dry_run:
            logger.setLevel(logging.DEBUG)

    args_keyvault = f'https://{args.keyvault}.vault.azure.net' if args.keyvault is not None else None

    env_file = Path(args.env_file)
    if args.subcommand == 'import' and not env_file.exists():
        if args.keyvault is None:
            raise ValueError(f'--vault-name not specified and environment configuration file {args.env_file} does not exist')

        _dict_to_yaml(env_file, {'configuration': {'keyvault': args_keyvault}}, indentation=2)

    environment, keyvault, configuration = _extract_metadata(args.env_file)

    if keyvault is None and args.keyvault is None:
        message = (
            'keyvault not specified in environment configuration file and no keyvault specified '
            'on the command line'
        )
        raise ValueError(message)

    keyvault = keyvault or args_keyvault

    if keyvault is None:
        raise ValueError('keyvault not specified, please specify a keyvault')

    client = get_keyvault_client(keyvault)

    try:
        if args.subcommand == 'import':
            return keyvault_import(client, environment, args, grizzly_context_root, configuration)
        elif args.subcommand == 'export':
            return keyvault_export(client, environment, args, grizzly_context_root, configuration)
        elif args.subcommand == 'diff':
            return diff(args.env_file, args.orig_file)
        else:
            raise ValueError(f'unknown subcommand {args.subcommand}')
    except ClientAuthenticationError:
        logger.error('authentication failed, if you are running from a resource which does not have a managed identity then you must run `az login` first.')
        return 1
    except ServiceRequestError:
        logger.error(f'{client.vault_url} does not resolve to an azure keyvault')
        return 1
    except ValueError:
        raise



import os

from typing import List, Dict, Any
from argparse import Namespace as Arguments

from . import register_parser
from .utils import (
    run_command,
)
from .run import create_parser as run_create_parser, run
from .argparse import ArgumentSubParser


@register_parser(order=2)
def create_parser(sub_parser: ArgumentSubParser) -> None:
    local_parser = sub_parser.add_parser('local', description='commands for running grizzly in local mode.')

    if local_parser.prog != 'grizzly-cli local':  # pragma: no cover
        local_parser.prog = 'grizzly-cli local'

    sub_parser = local_parser.add_subparsers(dest='subcommand')

    run_create_parser(sub_parser, parent='local')


def local(args: Arguments) -> int:
    if args.subcommand == 'run':
        return run(args, local_run)
    else:
        raise ValueError(f'unknown subcommand {args.subcommand}')


def local_run(args: Arguments, environ: Dict[str, Any], run_arguments: Dict[str, List[str]]) -> int:
    for key, value in environ.items():
        if key not in os.environ:
            os.environ[key] = value

    command = [
        'behave',
    ]

    if args.file is not None:
        command += [args.file]

    if len(run_arguments.get('master', [])) > 0 or len(run_arguments.get('worker', [])) > 0 or len(run_arguments.get('common', [])) > 0:
        command += run_arguments['master'] + run_arguments['worker'] + run_arguments['common']

    return run_command(command).return_code



from __future__ import annotations

import sys
import os

from typing import (
    List,
    Dict,
    Any,
    Callable,
    TextIO,
    cast,
)
from argparse import Namespace as Arguments
from platform import node as get_hostname
from datetime import datetime
from pathlib import Path
from contextlib import suppress
from jinja2 import Environment

import grizzly_cli
from .utils import (
    logger,
    find_variable_names_in_questions,
    ask_yes_no, get_input,
    distribution_of_users_per_scenario,
    requirements,
    find_metadata_notices,
    parse_feature_file,
    rm_rf,
)
from .utils.configuration import ScenarioTag, load_configuration, get_context_root
from .argparse import ArgumentSubParser
from .argparse.bashcompletion import BashCompletionTypes


def create_parser(sub_parser: ArgumentSubParser, parent: str) -> None:
    # grizzly-cli ... run ...
    run_parser = sub_parser.add_parser('run', description='execute load test scenarios specified in a feature file.')
    run_parser.add_argument(
        '--verbose',
        action='store_true',
        required=False,
        help=(
            'changes the log level to `DEBUG`, regardless of what it says in the feature file. gives more verbose logging '
            'that can be useful when troubleshooting a problem with a scenario.'
        )
    )
    run_parser.add_argument(
        '-T', '--testdata-variable',
        action='append',
        type=str,
        required=False,
        help=(
            'specified in the format `<name>=<value>`. avoids being asked for an initial value for a scenario variable.'
        )
    )
    run_parser.add_argument(
        '-y', '--yes',
        action='store_true',
        default=False,
        required=False,
        help='answer yes on any questions that would require confirmation',
    )
    run_parser.add_argument(
        '-e', '--environment-file',
        type=BashCompletionTypes.File('*.yaml', '*.yml'),
        required=False,
        default=None,
        help='configuration file with [environment specific information](/grizzly/framework/usage/variables/environment-configuration/)',
    )
    run_parser.add_argument(
        '--csv-prefix',
        nargs='?',
        const=True,
        default=None,
        help='write log statistics to CSV files with specified prefix, if no value is specified the description of the gherkin Feature tag will be used, suffixed with timestamp',
    )
    run_parser.add_argument(
        '--csv-interval',
        type=int,
        default=None,
        required=False,
        help='interval that statistics is collected for CSV files, can only be used in combination with `--csv-prefix`',
    )
    run_parser.add_argument(
        '--csv-flush-interval',
        type=int,
        default=None,
        required=False,
        help='interval that CSV statistics is flushed to disk, can only be used in combination with `--csv-prefix`',
    )
    run_parser.add_argument(
        '-l', '--log-file',
        type=str,
        default=None,
        required=False,
        help='save all `grizzly-cli` run output in specified log file',
    )
    run_parser.add_argument(
        '--log-dir',
        type=str,
        default=None,
        required=False,
        help='log directory suffix (relative to `requests/logs`) to save log files generated in a scenario',
    )
    run_parser.add_argument(
        '--dump',
        nargs='?',
        default=None,
        const=True,
        help=(
            'Dump parsed contents of file, can be useful when including scenarios from other feature files. If no argument is specified it '
            'will be dumped to stdout, the argument is treated as a filename'
        ),
    )
    run_parser.add_argument(
        '--dry-run',
        action='store_true',
        required=False,
        help='Will setup and run anything up until when locust should start. Useful for debugging feature files when developing new tests',
    )
    run_parser.add_argument(
        'file',
        nargs='+',
        type=BashCompletionTypes.File('*.feature'),
        help='path to feature file with one or more scenarios',

    )

    if run_parser.prog != f'grizzly-cli {parent} run':  # pragma: no cover
        run_parser.prog = f'grizzly-cli {parent} run'


@requirements(grizzly_cli.EXECUTION_CONTEXT)
def run(args: Arguments, run_func: Callable[[Arguments, Dict[str, Any], Dict[str, List[str]]], int]) -> int:
    # always set hostname of host where grizzly-cli was executed, could be useful
    environ: Dict[str, Any] = {
        'GRIZZLY_CLI_HOST': get_hostname(),
        'GRIZZLY_EXECUTION_CONTEXT': grizzly_cli.EXECUTION_CONTEXT,
        'GRIZZLY_MOUNT_CONTEXT': grizzly_cli.MOUNT_CONTEXT,
    }

    environment = Environment(autoescape=False, extensions=[ScenarioTag])
    feature_file = Path(args.file)
    environment_lock_file: str | None = None

    # during execution, create a temporary .lock.feature file that will be removed when done
    original_feature_lines = feature_file.read_text().splitlines()
    feature_lock_file = feature_file.parent / f'{feature_file.stem}.lock{feature_file.suffix}'

    try:
        buffer: List[str] = []
        remove_endif = False

        # remove if-statements containing variables (`{$ .. $}`)
        for line in original_feature_lines:
            stripped_line = line.strip()

            if stripped_line[:2] == '{%' and stripped_line[-2:] == '%}':
                if '{$' in stripped_line and '$}' in stripped_line and 'if' in stripped_line:
                    remove_endif = True
                    continue

                if remove_endif and 'endif' in stripped_line:
                    remove_endif = False
                    continue

            buffer.append(line)

        original_feature_content = '\n'.join(buffer)

        template = environment.from_string(original_feature_content)
        environment.extend(feature_file=feature_file, ignore_errors=False)
        feature_content = template.render()
        feature_lock_file.write_text(feature_content)

        if args.dump:
            output: TextIO
            if isinstance(args.dump, str):
                output = Path(args.dump).open('w+')
            else:
                output = sys.stdout

            print(feature_content, file=output)

            return 0

        args.file = feature_lock_file.as_posix()

        variables = find_variable_names_in_questions(args.file)
        questions = len(variables)
        manual_input = False

        if questions > 0 and not getattr(args, 'validate_config', False):
            logger.info(f'feature file requires values for {questions} variables')

            for variable in variables:
                name = f'TESTDATA_VARIABLE_{variable}'
                value = os.environ.get(name, '')
                while len(value) < 1:
                    value = get_input(f'initial value for "{variable}": ')
                    manual_input = True

                environ[name] = value

            logger.info('the following values was provided:')
            for key, value in environ.items():
                if not key.startswith('TESTDATA_VARIABLE_'):
                    continue
                logger.info(f'{key.replace("TESTDATA_VARIABLE_", "")} = {value}')

            if manual_input:
                ask_yes_no('continue?')

        notices = find_metadata_notices(args.file)

        if len(notices) > 0:
            if args.yes:
                output_func = cast(Callable[[str], None], logger.info)
            else:
                output_func = ask_yes_no

            for notice in notices:
                output_func(notice)

        if args.environment_file is not None:
            environment_file = os.path.realpath(args.environment_file)
            environment_lock_file = load_configuration(environment_file)
            environ.update({'GRIZZLY_CONFIGURATION_FILE': environment_lock_file})

        if args.dry_run:
            environ.update({'GRIZZLY_DRY_RUN': 'true'})

        if args.log_dir is not None:
            environ.update({'GRIZZLY_LOG_DIR': args.log_dir})

        if not getattr(args, 'validate_config', False):
            distribution_of_users_per_scenario(args, environ)

        run_arguments: Dict[str, List[str]] = {
            'master': [],
            'worker': [],
            'common': [],
        }

        if args.verbose:
            run_arguments['common'] += ['--verbose', '--no-logcapture', '--no-capture', '--no-capture-stderr']

        if args.csv_prefix is not None:
            if args.csv_prefix is True:
                parse_feature_file(args.file)
                if grizzly_cli.FEATURE_DESCRIPTION is None:
                    raise ValueError('feature file does not seem to have a `Feature:` description to use as --csv-prefix')

                csv_prefix = grizzly_cli.FEATURE_DESCRIPTION.replace(' ', '_')
                timestamp = datetime.now().astimezone().strftime('%Y%m%dT%H%M%S')
                setattr(args, 'csv_prefix', f'{csv_prefix}_{timestamp}')

            run_arguments['common'] += [f'-Dcsv-prefix="{args.csv_prefix}"']

            if args.csv_interval is not None:
                run_arguments['common'] += [f'-Dcsv-interval={args.csv_interval}']

            if args.csv_flush_interval is not None:
                run_arguments['common'] += [f'-Dcsv-flush-interval={args.csv_flush_interval}']

        return run_func(args, environ, run_arguments)
    finally:
        if environment_lock_file is not None:
            Path(environment_lock_file).unlink(missing_ok=True)

        feature_lock_file.unlink(missing_ok=True)

        with suppress(FileNotFoundError, ValueError):
            tmp_files = get_context_root() / 'files'
            rm_rf(tmp_files)



from typing import Generator
from argparse import Namespace as Arguments
from os import path
from pathlib import Path
from packaging.version import Version

from .utils import ask_yes_no
from .argparse import ArgumentSubParser
from . import EXECUTION_CONTEXT, register_parser

# prefix components:
space = '    '
branch = '│   '
# pointers:
tee = '├── '
last = '└── '


@register_parser(order=1)
def create_parser(sub_parser: ArgumentSubParser) -> None:
    # grizzly-cli init
    init_parser = sub_parser.add_parser('init', description=(
        'create a skeleton project with required structure and files.'
    ))

    init_parser.add_argument(
        'project',
        nargs=None,
        type=str,
        help='project name, a directory will be created with this name',
    )

    init_parser.add_argument(
        '--grizzly-version',
        type=str,
        required=False,
        default=None,
        help='specify which grizzly version to use for project, default is latest'
    )

    init_parser.add_argument(
        '--with-mq',
        action='store_true',
        default=False,
        required=False,
        help='if grizzly should be installed with IBM MQ support (external dependencies excluded)',
    )

    init_parser.add_argument(
        '-y', '--yes',
        action='store_true',
        default=False,
        required=False,
        help='automagically answer yes on any questions',
    )

    if init_parser.prog != 'grizzly-cli init':  # pragma: no cover
        init_parser.prog = 'grizzly-cli init'


def tree(dir_path: Path, prefix: str = '') -> Generator[str, None, None]:
    '''A recursive generator, given a directory Path object
    will yield a visual tree structure line by line
    with each line prefixed by the same characters

    credit: https://stackoverflow.com/a/59109706
    '''
    contents = sorted(list(dir_path.iterdir()))
    # contents each get pointers that are ├── with a final └── :
    pointers = [tee] * (len(contents) - 1) + [last]
    for pointer, sub_path in zip(pointers, contents):
        yield prefix + pointer + sub_path.name
        if sub_path.is_dir():  # extend the prefix and recurse:
            extension = branch if pointer == tee else space
            # i.e. space because last, └── , above so no more |
            yield from tree(sub_path, prefix=prefix + extension)


def init(args: Arguments) -> int:
    if path.exists(path.join(EXECUTION_CONTEXT, args.project)):
        print(f'"{args.project}" already exists in {EXECUTION_CONTEXT}')
        return 1

    if all([path.exists(path.join(EXECUTION_CONTEXT, p)) for p in ['environments', 'features', 'requirements.txt']]):
        print('oops, looks like you are already in a grizzly project directory', end='\n\n')
        print(EXECUTION_CONTEXT)
        for line in tree(Path(EXECUTION_CONTEXT)):
            print(line)
        return 1

    layout = f'''
    {args.project}
    ├── environments
    │   └── {args.project}.yaml
    ├── features
    │   ├── environment.py
    │   ├── steps
    │   │   └── steps.py
    │   ├── {args.project}.feature
    │   └── requests
    └── requirements.txt
'''

    message = f'the following structure will be created:\n{layout}'

    if not args.yes:
        ask_yes_no(f'{message}\ndo you want to create grizzly project "{args.project}"?')
    else:
        print(message)

    # create project root
    structure = Path(path.join(EXECUTION_CONTEXT, args.project))
    structure.mkdir()

    # create requirements.txt
    grizzly_dependency = 'grizzly-loadtester'

    if args.with_mq:
        grizzly_dependency = f'{grizzly_dependency}[mq]'

    if args.grizzly_version is not None:
        grizzly_dependency = f'{grizzly_dependency}=={args.grizzly_version}'

    (structure / 'requirements.txt').write_text(f'{grizzly_dependency}\n')

    # create environments/
    structure_environments = structure / 'environments'
    structure_environments.mkdir()

    # create environments/<project>.yaml
    (structure_environments / f'{args.project}.yaml').write_text('''configuration:
  template:
    host: https://localhost
''')

    # create features/ directory
    structure_features = structure / 'features'
    structure_features.mkdir()

    # create features/<project>.feature
    (structure_features / f'{args.project}.feature').write_text('''Feature: Template feature file
  Scenario: Template scenario
    Given a user of type "RestApi" with weight "1" load testing "$conf::template.host"
''')

    # create features/environment.py
    if args.grizzly_version is not None:
        version = Version(args.grizzly_version)
        if version < Version('2.6.0'):
            grizzly_behave_module = 'environment'
    else:
        grizzly_behave_module = 'behave'
    (structure_features / 'environment.py').write_text(f'from grizzly.{grizzly_behave_module} import *\n\n')

    # create features/requests directory
    (structure_features / 'requests').mkdir()

    # create features/steps directory
    structure_feature_steps = structure_features / 'steps'
    structure_feature_steps.mkdir()

    # create features/steps/steps.py
    (structure_feature_steps / 'steps.py').write_text('from grizzly.steps import *\n\n')

    print(f'successfully created project "{args.project}", with the following options:')
    print(f'{" " * 2}\u2022 {"with" if args.with_mq else "without"} IBM MQ support')
    if args.grizzly_version is not None:
        print(f'{" " * 2}\u2022 pinned to grizzly version {args.grizzly_version}')
    else:
        print(f'{" " * 2}\u2022 latest grizzly version')

    return 0



from os import environ
from tempfile import NamedTemporaryFile
from argparse import Namespace as Arguments
from getpass import getuser
from shutil import get_terminal_size

from grizzly_cli.argparse import ArgumentSubParser
from grizzly_cli.utils import run_command
from grizzly_cli import PROJECT_NAME, STATIC_CONTEXT


def create_parser(sub_parser: ArgumentSubParser) -> None:
    # grizzly-cli dist clean ...
    clean_parser = sub_parser.add_parser('clean', description=(
        'clean all grizzly compose project resources; containers, images, networks and volumes'
    ),)

    clean_parser.add_argument(
        '--no-images',
        dest='images',
        action='store_false',
        required=False,
        default=True,
        help='do not remove images',
    )

    clean_parser.add_argument(
        '--no-networks',
        dest='networks',
        action='store_false',
        required=False,
        default=True,
        help='do not remove networks',
    )

    if clean_parser.prog != 'grizzly-cli dist clean':  # pragma: no cover
        clean_parser.prog = 'grizzly-cli dist clean'


def clean(args: Arguments) -> int:
    suffix = '' if args.id is None else f'-{args.id}'
    tag = getuser()

    if args.project_name is not None:
        project_name = args.project_name
    else:
        project_name = PROJECT_NAME

    columns, lines = get_terminal_size()
    env = environ.copy()

    with NamedTemporaryFile() as fd:
        env.update({
            'GRIZZLY_PROJECT_NAME': project_name,
            'GRIZZLY_USER_TAG': tag,
            'GRIZZLY_CONTAINER_TTY': 'false',
            'GRIZZLY_LIMIT_NOFILE': '1024',
            'GRIZZLY_ENVIRONMENT_FILE': fd.name,
            'COLUMNS': str(columns),
            'LINES': str(lines),
        })

        compose_command = [
            args.container_system, 'compose',
            '-f', f'{STATIC_CONTEXT}/compose.yaml',
            '-p', f'{project_name}{suffix}-{tag}',
            'rm', '-f', '-s', '-v',
        ]

        result = run_command(compose_command, env=env)

    if args.images:
        command = [
            args.container_system,
            'image', 'rm', f'{project_name}:{tag}'
        ]

        run_command(command)

    if args.networks:
        command = [
            args.container_system,
            'network', 'rm', f'{project_name}{suffix}-{tag}_default',
        ]

        run_command(command)

    return result.return_code



import os
import sys
import argparse
import subprocess

from typing import List, Dict, Any, cast
from tempfile import NamedTemporaryFile
from getpass import getuser
from shutil import get_terminal_size
from argparse import Namespace as Arguments
from socket import gethostname
from json import loads as jsonloads
from io import StringIO
from pathlib import Path

from grizzly_cli import EXECUTION_CONTEXT, STATIC_CONTEXT, MOUNT_CONTEXT, PROJECT_NAME, register_parser
from grizzly_cli.utils import (
    run_command,
    get_default_mtu,
    list_images,
)
from grizzly_cli.run import create_parser as run_create_parser, run
from grizzly_cli.argparse import ArgumentSubParser
from .build import build as do_build, create_parser as build_create_parser
from .clean import clean as do_clean, create_parser as clean_create_parser


@register_parser(order=3)
def create_parser(sub_parser: ArgumentSubParser) -> None:
    dist_parser = sub_parser.add_parser('dist', description='commands for running grizzly i distributed mode.')

    dist_parser.add_argument(
        '--workers',
        type=int,
        required=False,
        default=1,
        help='how many instances of the `workers` container that should be created',
    )
    dist_parser.add_argument(
        '--container-system',
        type=str,
        choices=['podman', 'docker', None],
        required=False,
        default=None,
        help=argparse.SUPPRESS,
    )
    dist_parser.add_argument(
        '--id',
        type=str,
        required=False,
        default=None,
        help='unique identifier suffixed to compose project, should be used when the same user needs to run more than one instance of `grizzly-cli`',
    )
    dist_parser.add_argument(
        '--limit-nofile',
        type=int,
        required=False,
        default=10001,
        help='set system limit "number of open files"',
    )
    dist_parser.add_argument(
        '--health-retries',
        type=int,
        required=False,
        default=3,
        help='set number of retries for health check of master container',
    )
    dist_parser.add_argument(
        '--health-timeout',
        type=int,
        required=False,
        default=3,
        help='set timeout in seconds for health check of master container',
    )
    dist_parser.add_argument(
        '--health-interval',
        type=int,
        required=False,
        default=5,
        help='set interval in seconds between health checks of master container',
    )
    dist_parser.add_argument(
        '--registry',
        type=str,
        default=None,
        required=False,
        help='push built image to this registry, if the registry has authentication you need to login first',
    )
    dist_parser.add_argument(
        '--tty',
        action='store_true',
        default=False,
        required=False,
        help='start containers with a TTY enabled',
    )
    dist_parser.add_argument(
        '--wait-for-worker',
        type=str,
        default=None,
        required=False,
        help=(
            'sets enviroment variable LOCUST_WAIT_FOR_WORKERS_REPORT_AFTER_RAMP_UP, which tells master to wait '
            'this amount of time for worker report'
        )
    )

    dist_parser.add_argument(
        '--project-name',
        type=str,
        default=None,
        help='override project name, which otherwise would be the name of the directory where command is executed in',
    )

    group_build = dist_parser.add_mutually_exclusive_group()
    group_build.add_argument(
        '--force-build',
        action='store_true',
        required=False,
        help='force rebuild the grizzly projects container image (no cache)',
    )
    group_build.add_argument(
        '--build',
        action='store_true',
        required=False,
        help='rebuild the grizzly projects container images (with cache)',
    )
    group_build.add_argument(
        '--validate-config',
        action='store_true',
        required=False,
        help='validate and print compose project file',
    )

    if dist_parser.prog != 'grizzly-cli dist':  # pragma: no cover
        dist_parser.prog = 'grizzly-cli dist'

    sub_parser = dist_parser.add_subparsers(dest='subcommand')

    build_create_parser(sub_parser)
    clean_create_parser(sub_parser)
    run_create_parser(sub_parser, parent='dist')


def distributed(args: Arguments) -> int:
    if args.subcommand == 'run':
        return run(args, distributed_run)
    elif args.subcommand == 'build':
        return do_build(args)
    elif args.subcommand == 'clean':
        return do_clean(args)
    else:
        raise ValueError(f'unknown subcommand {args.subcommand}')


def distributed_run(args: Arguments, environ: Dict[str, Any], run_arguments: Dict[str, List[str]]) -> int:
    suffix = '' if args.id is None else f'-{args.id}'
    tag = getuser()

    if args.project_name is None:
        project_name = PROJECT_NAME
    else:
        project_name = args.project_name

    # default locust project
    compose_args: List[str] = [
        '-p', f'{project_name}{suffix}-{tag}',
        '-f', f'{STATIC_CONTEXT}/compose.yaml',
    ]

    if args.file is not None:
        os.environ['GRIZZLY_RUN_FILE'] = args.file

    if args.wait_for_worker is not None:
        os.environ['LOCUST_WAIT_FOR_WORKERS_REPORT_AFTER_RAMP_UP'] = f'{args.wait_for_worker}'

    mtu = get_default_mtu(args)

    if mtu is None and os.environ.get('GRIZZLY_MTU', None) is None:
        print('!! unable to determine MTU, try manually setting GRIZZLY_MTU environment variable if anything other than 1500 is needed')
        mtu = '1500'

    columns, lines = get_terminal_size()

    # set environment variables needed by compose files, when * compose executes
    os.environ['GRIZZLY_MTU'] = cast(str, mtu)
    os.environ['GRIZZLY_EXECUTION_CONTEXT'] = EXECUTION_CONTEXT
    os.environ['GRIZZLY_STATIC_CONTEXT'] = STATIC_CONTEXT
    os.environ['GRIZZLY_MOUNT_CONTEXT'] = MOUNT_CONTEXT
    os.environ['GRIZZLY_PROJECT_NAME'] = project_name
    os.environ['GRIZZLY_USER_TAG'] = tag
    os.environ['GRIZZLY_EXPECTED_WORKERS'] = str(args.workers)
    os.environ['GRIZZLY_LIMIT_NOFILE'] = str(args.limit_nofile)
    os.environ['GRIZZLY_HEALTH_CHECK_RETRIES'] = str(args.health_retries)
    os.environ['GRIZZLY_HEALTH_CHECK_INTERVAL'] = str(args.health_interval)
    os.environ['GRIZZLY_HEALTH_CHECK_TIMEOUT'] = str(args.health_timeout)
    os.environ['GRIZZLY_IMAGE_REGISTRY'] = getattr(args, 'registry', None) or ''
    os.environ['GRIZZLY_CONTAINER_TTY'] = 'true' if args.tty else 'false'
    os.environ['COLUMNS'] = str(columns)
    os.environ['LINES'] = str(lines)

    grizzly_mount_context_path = ''

    name_template = '{project}{suffix}-{tag}-{node}-{index}'

    if EXECUTION_CONTEXT != MOUNT_CONTEXT:
        hostname = gethostname()
        output = subprocess.check_output(
            [args.container_system, 'container', 'inspect', '-f', '{{ json .Mounts }}', hostname],
            encoding='utf-8',
        )
        container_mounts = jsonloads(output)
        for container_mount in container_mounts:
            if container_mount['Source'] != MOUNT_CONTEXT:
                continue

            grizzly_mount_context_path = EXECUTION_CONTEXT.replace(container_mount['Destination'], '')[1:]
            break

    os.environ['GRIZZLY_MOUNT_PATH'] = grizzly_mount_context_path

    if len(run_arguments.get('master', [])) > 0:
        os.environ['GRIZZLY_MASTER_RUN_ARGS'] = ' '.join(run_arguments['master'])

    if len(run_arguments.get('worker', [])) > 0:
        os.environ['GRIZZLY_WORKER_RUN_ARGS'] = ' '.join(run_arguments['worker'])

    if len(run_arguments.get('common', [])) > 0:
        os.environ['GRIZZLY_COMMON_RUN_ARGS'] = ' '.join(run_arguments['common'])

    # check if we need to build image
    images = list_images(args)

    with NamedTemporaryFile() as fd:
        # file will be deleted when conContainertext exits
        if len(environ) > 0:
            for key, value in environ.items():
                if key == 'GRIZZLY_CONFIGURATION_FILE':
                    value = value.replace(EXECUTION_CONTEXT, MOUNT_CONTEXT).replace(MOUNT_CONTEXT, '/srv/grizzly')

                fd.write(f'{key}={value}\n'.encode('utf-8'))

        fd.write(f'COLUMNS={columns}\n'.encode('utf-8'))
        fd.write(f'LINES={lines}\n'.encode('utf-8'))
        fd.write(f'GRIZZLY_CONTAINER_TTY={os.environ["GRIZZLY_CONTAINER_TTY"]}\n'.encode('utf-8'))

        if args.wait_for_worker is not None:
            fd.write(f'LOCUST_WAIT_FOR_WORKERS_REPORT_AFTER_RAMP_UP="{args.wait_for_worker}"'.encode('utf-8'))

        fd.flush()

        os.environ['GRIZZLY_ENVIRONMENT_FILE'] = fd.name

        validate_config = getattr(args, 'validate_config', False)

        compose_command = [
            args.container_system, 'compose',
            *compose_args,
            'config',
        ]

        result = run_command(compose_command, silent=not validate_config)

        if validate_config or result.return_code != 0:
            if result.return_code != 0 and not validate_config:
                print('!! something in the compose project is not valid, check with:')
                argv = sys.argv[:]
                argv.insert(argv.index('dist') + 1, '--validate-config')
                print(f'grizzly-cli {" ".join(argv[1:])}')

            return result.return_code

        if images.get(project_name, {}).get(tag, None) is None or args.force_build or args.build:
            rc = do_build(args)
            if rc != 0:
                print(f'!! failed to build {project_name}, rc={rc}')
                return rc

        compose_scale_argument = ['--scale', f'worker={args.workers}']

        # bring up containers
        compose_command = [
            args.container_system, 'compose',
            *compose_args,
            'up',
            *compose_scale_argument,
            '--remove-orphans',
        ]

        result = run_command(compose_command, verbose=args.verbose)

        try:
            output = subprocess.check_output(
                [
                    args.container_system,
                    'inspect',
                    '-f',
                    '{{ .State.ExitCode }}',
                    name_template.format(
                        project=project_name,
                        suffix=suffix,
                        tag=tag,
                        node='master',
                        index=1,
                    ),
                ],
                encoding='utf-8',
            )
            result.return_code = int(output.strip())
        except:
            result.return_code = 1

        # stop containers
        compose_command = [
            args.container_system, 'compose',
            *compose_args,
            'stop',
        ]

        run_command(compose_command)

        if result.return_code != 0:
            if result.abort_timestamp is not None:
                master_node_name = name_template.format(
                    project=project_name,
                    suffix=suffix,
                    tag=tag,
                    node='master',
                    index=1,
                )

                since_timestamp = result.abort_timestamp.strftime('%Y-%m-%dT%H:%M:%SZ')
                command = [args.container_system, 'container', 'logs', '--since', since_timestamp, master_node_name]

                missed_output = subprocess.check_output(
                    command,
                    encoding='utf-8',
                    shell=False,
                    universal_newlines=True,
                    stderr=subprocess.STDOUT,
                ).split('\n')

                log_file = Path(args.log_file).open('a+') if args.log_file is not None else StringIO()

                try:
                    for line in missed_output:
                        formatted_line = f'{master_node_name}  | {line}'
                        print(formatted_line)
                        log_file.write(f'{formatted_line}\n')
                except:
                    pass
                finally:
                    log_file.close()

            print('\n!! something went wrong, check full container logs with:')
            template = '{container_system} container logs {name_template}'
            print(template.format(
                container_system=args.container_system,
                name_template=name_template.format(
                    project=project_name,
                    suffix=suffix,
                    tag=tag,
                    node='master',
                    index=1,
                ),
            ))

            for worker in range(1, args.workers + 1):
                print(template.format(
                    container_system=args.container_system,
                    name_template=name_template.format(
                        project=project_name,
                        suffix=suffix,
                        tag=tag,
                        node='worker',
                        index=worker,
                    ),
                ))

        return result.return_code
